############################################################################
#                                                                          #
# Copyright (c) 2019-2020 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Tests creating several chained datasets in one job.
Exercises DatasetWriter.finish and the chaining logic
including callbacks with SkipDataset.
'''

from accelerator.dataset import DatasetWriter, SkipDataset
from accelerator.extras import DotDict

def prepare(params):
	dws = {}
	prev = None
	for name in "abcdefgh":
		dw = DatasetWriter(name=name, previous=prev)
		dw.add("ds", "ascii")
		dw.add("num", "number")
		dws[name] = dw
		prev = dw
	return dws

def analysis(sliceno, prepare_res):
	# First slice goes down, datasets c - h contain negative numbers.
	if sliceno == 0:
		value = 3
		add = -2
	else:
		value = sliceno
		add = sliceno
	for name, dw in sorted(prepare_res.items()):
		dw.write(name, value)
		value += add

def synthesis(prepare_res, params):
	ds = DotDict()
	# Must be finished in order (.previous must be finished when .finish is called.)
	for name, dw in sorted(prepare_res.items()):
		ds[name] = dw.finish()
	last = ds.h
	assert last.chain() == sorted(ds.values())
	ds.last = last
	test_partial_chains(ds)
	test_filters(ds)

def test_partial_chains(ds):
	alles = list(ds.last.iterate_chain(None))
	part1 = list(ds.b.iterate_chain(None))
	part2 = list(ds.e.iterate_chain(None, length=3))
	def only_f_and_g_cb(d):
		if d[-2:] not in ("/f", "/g"):
			raise SkipDataset
	seen = []
	def record_cb(d):
		seen.append(d)
	part3 = list(ds.g.iterate_chain(None, pre_callback=only_f_and_g_cb, post_callback=record_cb))
	assert seen == ds.g.chain(length=2)
	part4 = list(ds.last.iterate(None))
	assert alles == part1 + part2 + part3 + part4
	two_by_length = ds.d.chain(length=2)
	two_by_id = ds.d.chain(stop_ds=ds.b)
	assert two_by_length == two_by_id
	def stop_on(name):
		def cb(d):
			if d == name:
				raise StopIteration
		return cb
	just_two = list(ds.last.iterate_list(None, None, [ds.a, ds.b]))
	two_by_pre_stop = list(ds.last.iterate_chain(None, pre_callback=stop_on(ds.c)))
	assert two_by_pre_stop == just_two
	two_by_post_stop = list(ds.last.iterate_chain(None, post_callback=stop_on(ds.b)))
	assert two_by_post_stop == just_two

def test_filters(ds):
	just_c = list(ds.c.iterate(None))
	filtered_to_c = list(ds.last.iterate_chain(None, filters={"ds": "c".__eq__}))
	assert just_c == filtered_to_c
	just_c_0 = list(ds.c.iterate(0))
	# Can't use (0).__gt__ in py2.
	just_c_filtered_negative = list(ds.c.iterate(None, filters={"num": lambda v: v < 0}))
	assert just_c_0 == just_c_filtered_negative
	# test two filters
	alles_filtered_c_negative = list(ds.last.iterate_chain(None, filters={"num": lambda v: v < 0, "ds": "c".__eq__}))
	assert just_c_0 == alles_filtered_c_negative
	# test single tuple filter
	alles_filtered_c_negative_again = list(ds.last.iterate_chain(None, filters=lambda t: t[0] == "c" and t[1] < 0))
	assert alles_filtered_c_negative == alles_filtered_c_negative_again
	# test filtering something translated
	# Can't use (2).__add__ in (some) py2.
	just_0_plus_2_pos = list(ds.last.iterate_chain(0, filters={"num": lambda v: v > 0}, translators={"num": lambda v: v + 2}))
	just_0 = list(ds.last.iterate_chain(0))
	just_0_plus_2_pos_manual = [(t[0], t[1] + 2,) for t in just_0 if t[1] + 2 > 0]
	assert just_0_plus_2_pos == just_0_plus_2_pos_manual
	# Test filtering with (non-tupled) single column
	just_0_num_plus_2_pos = list(ds.last.iterate_chain(0, columns="num", filters={"num": lambda v: v > 0}, translators={"num": lambda v: v + 2}))
	assert just_0_num_plus_2_pos == [t[1] for t in just_0_plus_2_pos]



############################################################################
#                                                                          #
# Copyright (c) 2021-2022 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test the "ax grep" shell command. This isn't testing complex regexes,
but rather the various output options and data types.
'''

options = dict(
	command_prefix=['ax', '--config', '/some/path/here'],
)

from subprocess import check_output, Popen
import datetime
import os
import errno
import json
from itertools import cycle
from functools import partial

from accelerator.compat import PY2, PY3, izip_longest
from accelerator.dsutil import _convfuncs

def grep_text(args, want, sep='\t', encoding='utf-8', unordered=False, check_output=check_output):
	if not unordered:
		args = ['--ordered'] + args
	cmd = options.command_prefix + ['grep'] + args
	res = check_output(cmd)
	res = res.split(b'\n')[:-1]
	if len(want) != len(res):
		raise Exception('%r gave %d lines, wanted %d.\n%r' % (cmd, len(res), len(want), res,))
	if encoding:
		res = [el.decode(encoding, 'replace') for el in res]
	typ = type(sep)
	want = [sep.join(typ(el) for el in l) for l in want]
	for lineno, (want, got) in enumerate(zip(want, res), 1):
		if want != got:
			raise Exception('%r gave wrong result on line %d:\nWant: %r\nGot:  %r' % (cmd, lineno, want, got,))

# like subprocess.check_output except stdout is a pty
def check_output_pty(cmd):
	a, b = os.openpty()
	p = Popen(cmd, stdout=a)
	os.close(a)
	res = []
	while not res or data:
		try:
			data = os.read(b, 1024)
		except OSError as e:
			# On Linux a pty will return
			# OSError: [Errno 5] Input/output error
			# instead of b'' for EOF. Don't know why.
			# Let's try to be a little restrictive in what we catch.
			if e.errno != errno.EIO:
				raise
			data = b''
		res.append(data)
	os.close(b)
	return b''.join(res)

def grep_json(args, want):
	cmd = options.command_prefix + ['grep', '--ordered', '--format=json'] + args
	res = check_output(cmd)
	res = res.decode('utf-8', 'surrogatepass')
	res = res.split('\n')[:-1]
	if len(want) != len(res):
		raise Exception('%r gave %d lines, wanted %d.\n%r' % (cmd, len(res), len(want), res,))
	for lineno, (want, got) in enumerate(zip(want, res), 1):
		try:
			got = json.loads(got)
		except Exception as e:
			raise Exception('%r made bad json %r on line %d: %s' % (cmd, got, lineno, e,))
		if want != got:
			raise Exception('%r gave wrong result on line %d:\nWant: %r\nGot:  %r' % (cmd, lineno, want, got,))

if PY2:
	def mk_bytes(low, high):
		return b''.join(chr(c) for c in range(low, high))
else:
	def mk_bytes(low, high):
		return bytes(range(low, high))

# looks like 'bar' when matching but 'foo' when printing
# intended to catch if objects are evaluated too many times
class TricksyObject:
	def __init__(self):
		self.counter = 0
	def __str__(self):
		self.counter += 1
		if self.counter == 1:
			return 'bar'
		elif self.counter == 2:
			return 'foo'
		else:
			return 'oops'

def synthesis(job, slices):
	used_types = set()
	def mk_ds(name, types, *lines, **kw):
		dw = job.datasetwriter(name=name, **kw)
		for t in types:
			if isinstance(t, tuple):
				t, name = t
			else:
				name = t
			dw.add(name, t)
			used_types.add(t)
		for sliceno, line in izip_longest(range(slices), lines):
			if sliceno is not None:
				dw.set_slice(sliceno)
			if line:
				dw.write(*line)
		return dw.finish()
	HDR_HI = '\x1b[94;1m'
	HDR_HI_POST = '\x1b[22;39m'
	SEP_HI = '\x1b[36;4m'
	SEP_HI_POST = '\x1b[24;39m'
	COMMA_HI = SEP_HI + ',' + SEP_HI_POST
	TAB_HI = SEP_HI + '\t' + SEP_HI_POST
	os.unsetenv('NO_COLOR')
	os.unsetenv('CLICOLOR')
	os.unsetenv('CLICOLOR_FORCE')
	os.putenv('XDG_CONFIG_HOME', job.path) # make sure we can't be messed up by config

	def frame(framevalues, *a):
		framevalues = cycle(framevalues)
		return [next(framevalues) + el + next(framevalues) for el in a]

	hdrframe = partial(frame, [HDR_HI, HDR_HI_POST])
	sepframe = partial(frame,  ['', '', SEP_HI, SEP_HI_POST])
	hdrsepframe = partial(frame,  [HDR_HI, HDR_HI_POST, SEP_HI, SEP_HI_POST])

	# start with testing basic output, chaining, column selection and headers.
	a = mk_ds('a', ['int32', 'int64'], [100, 200], [101, 201])
	b = mk_ds('b', ['int32'], [1000], [1001], previous=a)
	c = mk_ds('c', ['float64', 'int32'], [1.42, 3], previous=b)
	grep_text(['', a], [[100, 200], [101, 201]])
	grep_json(['', a], [{'int32': 100, 'int64': 200}, {'int32': 101, 'int64': 201}])
	grep_text(['-S', '', a], [[0, 100, 200], [1, 101, 201]])
	grep_text(['', b + '~'], [[100, 200], [101, 201]]) # verify ds parsing happens
	grep_json(['-S', '-L', '-D', '-c', '', b], [
		{'dataset': a, 'sliceno': 0, 'lineno': 0, 'data': {'int32': 100, 'int64': 200}},
		{'dataset': a, 'sliceno': 1, 'lineno': 0, 'data': {'int32': 101, 'int64': 201}},
		{'dataset': b, 'sliceno': 0, 'lineno': 0, 'data': {'int32': 1000}},
		{'dataset': b, 'sliceno': 1, 'lineno': 0, 'data': {'int32': 1001}},
	])
	grep_text(['-t', '2', '', a], [[100, '"200"'], [101, '"201"']], sep='2') # stupid separator leads to escaping
	grep_text(['-t', '2', '-f', 'raw', '', a], [[100, 200], [101, 201]], sep='2') # but not in the raw format

	# missing columns
	grep_text(['-M', '-c', '', c, 'int32'], [[100], [101], [1000], [1001], [3]]) # not anything missing
	grep_text(['-M', '-c', '-g', 'int64', '', c], [[100, 200], [101, 201]])
	grep_text(['-M', '-c', '-H', '', c, 'int32', 'float64'], [['int32'], [100], [101], [1000], [1001], ['int32', 'float64'], [3, 1.42]])
	grep_text(['-M', '-c', '-H', '', c, 'non-existant'], [])
	grep_text(['-M', '-c', '-g', 'float64', '-g', 'int64', '-H', '', c], [['int32', 'int64'], [100, 200], [101, 201], ['float64', 'int32'], [1.42, 3]])
	grep_text(['-M', '-c', '-H', '', c, 'int64', 'float64'], [['int64'], [200], [201], ['float64'], [1.42]])

	# context
	ctx = job.datasetwriter(name='ctx', columns={'a': 'int32'}, allow_missing_slices=True)
	ctx.set_slice(0)
	for v in range(100):
		if v == 32:
			ctx.set_slice(1)
		ctx.write(v)
	ctx = ctx.finish()
	ctx2 = job.datasetwriter(name='ctx2', columns={'a': 'ascii', 'b': 'ascii'}, allow_missing_slices=True, previous=ctx)
	ctx2.set_slice(1)
	for v in range(26):
		ctx2.write(chr(65 + v), chr(97 + v))
	ctx2 = ctx2.finish()
	# only one before 33 because of slicing, only one after 98 because the ds ends (even though ctx2 continues in slice 1)
	grep_text(['-c', '-A', '3', '-B', '2', '(33|98|Z)', ctx2], [[32], [33], [34], [35], [36], [96], [97], [98], [99], ['X', 'x'], ['Y', 'y'], ['Z', 'z']])

	# --invert-match
	grep_text(['-v', '[a-mO-Y]', ctx2], [[ch.upper(), ch] for ch in 'nz'])
	# ... with --context
	grep_text(['-C', '1', '-v', '[a-mO-Y]', ctx2], [[ch.upper(), ch] for ch in 'mnoyz'])

	# several (partially overlapping) patterns
	grep_text(['-e', '[moyz]', '-e', '[mNO]', ctx2], [[ch.upper(), ch] for ch in 'mnoyz'])
	# ... with --invert-match
	grep_text(['-v', '-e', '[a-ls-x]', '-e', '[p-v]', '-e', 'z', ctx2], [[ch.upper(), ch] for ch in 'mnoy'])
	# ... and --context
	grep_text(['-C', '1', '-v', '-e', '[a-ns-x]', '-e', '[p-v]', '-e', 'y', ctx2], [[ch.upper(), ch] for ch in 'nopyz'])

	# several datasets and columns (testing that unqualified and qualified works the same)
	grep_text(['-M', '01', a, b, 'int32', 'int64'], [[101, 201], [1001]])
	grep_text(['-M', '-d', a, '01', b, 'int32', 'int64'], [[101, 201], [1001]])
	grep_text(['-M', a, b, 'int32', 'int64', '-e', '01'], [[101, 201], [1001]])
	grep_text(['-M', '-d', a, '--column', 'int32', '-n', 'int64', '--regexp', '01', '--dataset', b], [[101, 201], [1001]])
	# + several patterns
	grep_text(['-M', '-e', '000', a, b, 'int32', 'int64', '-e', '01'], [[101, 201], [1000], [1001]])

	# try some colour
	grep_text(['--colour', '-t', ',', '-D', '-S', '-H', '', a], [hdrframe('[DATASET]', '[SLICE]', 'int32', 'int64'), [a, 0, 100, 200], [a, 1, 101, 201]], sep=COMMA_HI)
	os.putenv('CLICOLOR_FORCE', '1')
	grep_text(['-t', ',', '-L', '-S', '-H', '', a], [hdrframe('[SLICE]', '[LINE]', 'int32', 'int64'), [0, 0, 100, 200], [1, 0, 101, 201]], sep=COMMA_HI)
	grep_text(['-t', ',', '-L', '-S', '-H', '--colour=never', '', a], [['[SLICE]', '[LINE]', 'int32', 'int64'], [0, 0, 100, 200], [1, 0, 101, 201]], sep=',')
	grep_text(['-t', ',', '-D', '-H', '', b], [hdrframe('[DATASET]', 'int32'), [b, 1000], [b, 1001]], sep=COMMA_HI)
	grep_text(['-t', ',', '-C', '1', 'F', ctx2], [['E', 'e'], ['\x1b[31mF\x1b[39m', 'f'], ['G', 'g']], sep=COMMA_HI)
	grep_text(['-t', ',', '-D', '-H', '-c', '', b], [hdrframe('[DATASET]', 'int32', 'int64'), [a, 100, 200], [a, 101, 201], hdrframe('[DATASET]', 'int32'), [b, 1000], [b, 1001]], sep=COMMA_HI)
	grep_json(['-s', '0', '', a], [{'int32': 100, 'int64': 200}]) # no colour in json
	grep_text(['-s', '0', '', b, a], [['1000'], ['100', '200']], sep=TAB_HI)
	grep_text(['--color=never', '0', b], [[1000], [1001]])
	os.unsetenv('CLICOLOR_FORCE')
	os.putenv('NO_COLOR', '')
	grep_text(['--colour', 'always', '0', b], [['1\x1b[31m000\x1b[39m'], ['1\x1b[31m00\x1b[39m1']], sep=TAB_HI)
	os.unsetenv('NO_COLOR')

	# test the tab-replacing separator handling
	grep_text(['--color=always', '--tab-length=8', '-H', '-S', '^1', a, c], [
			hdrsepframe('[SLICE]', ' ', 'int32', '   ', 'int64'),
			sepframe('0', '       ', '\x1b[31m1\x1b[39m00', '     ', '200'),
			sepframe('1', '       ', '\x1b[31m1\x1b[39m01', '     ', '201'),
			hdrsepframe('[SLICE]', ' ', 'float64', ' ', 'int32'),
			sepframe('0', '       ', '\x1b[31m1\x1b[39m.42', '    ', '3'),
		], sep='',
	)
	# different length
	grep_text(['--color=always', '--tab-length=3', '-H', '-S', '^1', a, c], [
			hdrsepframe('[SLICE]', '  ', 'int32', ' ', 'int64'),
			sepframe('0', '  ', '\x1b[31m1\x1b[39m00', '   ', '200'),
			sepframe('1', '  ', '\x1b[31m1\x1b[39m01', '   ', '201'),
			hdrsepframe('[SLICE]', '  ', 'float64', '  ', 'int32'),
			sepframe('0', '  ', '\x1b[31m1\x1b[39m.42', '  ', '3'),
		], sep='',
	)
	# with a PTY, to see that this defaults to colour and smart expanded tabs
	grep_text(['-S', '^1', c, 'int32', 'float64'], [
			sepframe('0', '               ', '3', '               ', '\x1b[31m1\x1b[39m.42'),
		], sep='', check_output=check_output_pty,
	)

	if PY3: # no pickle type on PY2
		pickle = mk_ds('pickle', ['pickle'], [TricksyObject()], [''], [{'foo'}])
		grep_text(['', pickle], [['foo'], [''], ["{'foo'}"]])
		grep_text(['.', pickle], [['foo'], ["{'foo'}"]])
		grep_text(['bar', pickle], [['foo']])
		# using -g with the same columns as output is a NOP
		grep_text(['-g', 'pickle', 'bar', pickle], [['foo']])
		# but using it with a different set of columns is not
		pickle2 = mk_ds('pickle2', ['ascii'], ['a'], ['b'], ['c'], parent=pickle)
		grep_text(['-g', 'pickle', 'bar', pickle2], [['a', 'bar']])
		# order doesn't matter for equality, so here we're back to double evaluation.
		grep_text(['-g', 'pickle', '-g', 'ascii', 'bar', pickle2], [['a', 'foo']])
		bytespickle = mk_ds('bytespickle', ['pickle'], [b'\xf0'], [b'\t'])
		# pickles are str()d, not special cased like bytes columns
		grep_text(['-f', 'raw', 'xf0', bytespickle], [["b'\\xf0'"]])
		grep_json(['', bytespickle], [{'pickle': "b'\\xf0'"}, {'pickle': "b'\\t'"}])

	# --only-matching, both the part (default) and columns (with -l) in both csv and json
	grep_text(['-o', '-c', '1', b], [['1', ''], ['11', '1'], ['1'], ['11']])
	grep_text(['-o', '-l', '-c', '1', b], [['100', ''], ['101', '201'], ['1000'], ['1001']])
	grep_json(['-o', '-c', '1', b], [{'int32': '1', 'int64': ''}, {'int32': '11', 'int64': '1'}, {'int32': '1'}, {'int32': '11'}])
	grep_json(['-o', '-l', '-c', '1', b], [{'int32': 100}, {'int32': 101, 'int64': 201}, {'int32': 1000}, {'int32': 1001}])

	# check all possible byte values in all output formats
	allbytes = mk_ds('allbytes', ['ascii', 'bytes'],
		['control chars', mk_bytes(0, 32)],
		['printable', mk_bytes(32, 128)],
		['not ascii', mk_bytes(128, 256)],
	)
	if PY2:
		not_ascii = '\ufffd'.encode('utf-8') * 128
	else:
		not_ascii = mk_bytes(128, 256)
	grep_text(
		['--format=raw', '', allbytes],
		[
			[b'control chars', mk_bytes(0, 10)],
			[mk_bytes(11, 32)], # we end up with an extra line because the control chars have a newline
			[b'printable', mk_bytes(32, 128)],
			[b'not ascii', not_ascii],
		],
		encoding=None,
		sep=b'\t',
	)
	if PY3:
		not_ascii = not_ascii.decode('utf-8', 'surrogateescape').encode('utf-8', 'surrogatepass')
	grep_text(
		['', allbytes],
		[
			[b'control chars', b'"' + mk_bytes(0, 10) + b'\\n' + mk_bytes(11, 32) + b'"'],
			[b'printable', mk_bytes(32, 128)],
			[b'not ascii', not_ascii],
		],
		encoding=None,
		sep=b'\t',
	)
	grep_json(['', allbytes], [
		{'ascii': 'control chars', 'bytes': mk_bytes(0, 32).decode('utf-8', 'surrogateescape' if PY3 else 'replace')},
		{'ascii': 'printable', 'bytes': mk_bytes(32, 128).decode('utf-8', 'surrogateescape' if PY3 else 'replace')},
		{'ascii': 'not ascii', 'bytes': mk_bytes(128, 256).decode('utf-8', 'surrogateescape' if PY3 else 'replace')},
	])

	# header printing should happen between datasets only when columns change,
	# and must wait for all slices for each switch.
	# to make this predictable without -O, only one slice is used per column set.
	columns = [
		('int32', 'int64',),
		('int64', 'int32',), # not actually a change
		('int32', 'number',),
		('int32',),
		('int32',),
		('int32',),
		('int64',),
	]
	values_every_time = range(10)
	previous = None
	previous_cols = []
	slice = 0
	header_test = []
	for ds_ix, cols in enumerate(columns):
		dw = job.datasetwriter(name='header test %d' % (ds_ix,), previous=previous, allow_missing_slices=True)
		for col in cols:
			dw.add(col, col)
		if sorted(cols) != previous_cols:
			# columns changed, so switch slice to make failure more likely
			previous_cols = sorted(cols)
			slice = (slice + 1) % slices
		dw.set_slice(slice)
		for value in values_every_time:
			args = (value,) * len(cols)
			dw.write(*args)
		previous = dw.finish()
		header_test.append((previous, slice))
	grep_text(
		['-H', '-c', '', previous],
			[['int32', 'int64']] +
			[(v, v,) for v in values_every_time] +
			[(v, v,) for v in values_every_time] +
			[['int32', 'number']] +
			[(v, v,) for v in values_every_time] +
			[['int32']] +
			[(v,) for v in values_every_time] +
			[(v,) for v in values_every_time] +
			[(v,) for v in values_every_time] +
			[['int64']] +
			[(v,) for v in values_every_time],
		unordered=True,
	)
	# test that only a single slice with header changes isn't a problem.
	grep_text(
		['-H', '-D', '-c', '-s', '2', '7', header_test[3][0]], [
			['[DATASET]', 'int32', 'int64'],
			['[DATASET]', 'int32', 'number'],
			[header_test[2][0], '7', '7'], # the other datasets don't have anything in slice 2
			['[DATASET]', 'int32'],
		],
	)
	# test that repeating datasets doesn't repeat header outputs
	grep_text(
		['-H', '-D', '-s', '2', '7', header_test[2][0], header_test[2][0], header_test[3][0], header_test[3][0]], [
			['[DATASET]', 'int32', 'number'],
			[header_test[2][0], '7', '7'],
			[header_test[2][0], '7', '7'],
			['[DATASET]', 'int32'],
		],
	)

	# test --chain-length and --stop-ds
	grep_text(
		['-H', '-D', '-c', '--chain-length=3', '7', header_test[3][0]], [
			['[DATASET]', 'int32', 'int64'],
			[header_test[1][0], '7', '7'],
			['[DATASET]', 'int32', 'number'],
			[header_test[2][0], '7', '7'],
			['[DATASET]', 'int32'],
			[header_test[3][0], '7'],
		],
	)
	# --stop-ds happens first and wins
	grep_text(
		['-H', '-D', '-c', '--chain-length=3', '--stop-ds', header_test[1][0], '7', header_test[3][0]], [
			['[DATASET]', 'int32', 'number'],
			[header_test[2][0], '7', '7'],
			['[DATASET]', 'int32'],
			[header_test[3][0], '7'],
		],
	)
	# ...with several values
	grep_text(
		['-H', '-D', '-c', '--chain-length=3', '--chain-length=2', '7', header_test[3][0], header_test[4][0]], [
			['[DATASET]', 'int32', 'int64'],
			[header_test[1][0], '7', '7'],
			['[DATASET]', 'int32', 'number'],
			[header_test[2][0], '7', '7'],
			['[DATASET]', 'int32'],
			[header_test[3][0], '7'],
			[header_test[3][0], '7'],
			[header_test[4][0], '7'],
		],
	)
	# same thing using --stop-ds for the second ds. (--chain-length=3 applies to both, but only matters for the first)
	grep_text(
		['-H', '-D', '-c', '--chain-length=3', '--stop-ds=', '--stop-ds', header_test[2][0], '7', header_test[3][0], header_test[4][0]], [
			['[DATASET]', 'int32', 'int64'],
			[header_test[1][0], '7', '7'],
			['[DATASET]', 'int32', 'number'],
			[header_test[2][0], '7', '7'],
			['[DATASET]', 'int32'],
			[header_test[3][0], '7'],
			[header_test[3][0], '7'],
			[header_test[4][0], '7'],
		],
	)

	# test --list-matching
	grep_text(['-l', '-c', '', previous], [[ds] for ds, _ in header_test])
	want = [[ds, str(sliceno)] for ds, sliceno in header_test]
	grep_text(['-l', '-c', '-S', '', previous], want)
	grep_text(['-l', '-c', '-S', '-H', '', previous], [['[DATASET]', '[SLICE]']] + want)
	grep_json(['-l', '-c', '', previous], [{'dataset': ds} for ds, _ in header_test])
	grep_json(['-l', '-c', '-S', '', previous], [{'dataset': ds, 'sliceno': sliceno} for ds, sliceno in header_test])
	# test escaping
	grep_text(['-l', '-t', '/', '-S', '', previous], [['"%s"/%d' % header_test[-1]]])

	# more escaping
	escapy = mk_ds('escapy',
		[('ascii', 'spaced name'), ('unicode', 'tabbed\tname')],
		['comma', 'foo,bar'],
		['tab', 'foo\tbar'],
		['newline', 'a brand new\nline'],
		['doublequote start', '"foo'],
		['doublequote inside', 'f"oo'],
		['doublequote end', 'foo"'],
		['singlequote start', "'foo"],
		['singlequote inside', "f'oo"],
		['singlequote end', "foo'"],
	)
	grep_text(['-H', '', escapy], [
		['spaced name', '"tabbed\tname"'],
		['comma', 'foo,bar'],
		['tab', '"foo\tbar"'],
		['newline', 'a brand new\\nline'],
		['doublequote start', '"""foo"'],
		['doublequote inside', 'f"oo'],
		['doublequote end', '"foo"""'],
		['singlequote start', "\"'foo\""],
		['singlequote inside', "f'oo"],
		['singlequote end', "\"foo'\""],
	])
	grep_text(['-H', '-f', 'raw', '', escapy], [
		['spaced name', 'tabbed\tname'],
		['comma', 'foo,bar'],
		['tab', 'foo\tbar'],
		['newline', 'a brand new'],
		['line'], # newline is not escaped
		['doublequote start', '"foo'],
		['doublequote inside', 'f"oo'],
		['doublequote end', 'foo"'],
		['singlequote start', "'foo"],
		['singlequote inside', "f'oo"],
		['singlequote end', "foo'"],
	])
	grep_text(['-H', '-t', ',', '(bar|newline|end)', escapy], [
		['spaced name', 'tabbed\tname'],
		['comma', '"foo,bar"'],
		['tab', 'foo\tbar'],
		['newline', 'a brand new\\nline'],
		['doublequote end', '"foo"""'],
		['singlequote end', "\"foo'\""],
	], sep=',')
	grep_text(['-H', '-t', ' ', '(tab|inside)', escapy], [
		['"spaced name"', 'tabbed\tname'],
		['tab', 'foo\tbar'],
		['"doublequote inside"', 'f"oo'],
		['"singlequote inside"', "f'oo"],
	], sep=' ')
	grep_json(['', escapy], [
		{'spaced name': 'comma', 'tabbed\tname': 'foo,bar'},
		{'spaced name': 'tab', 'tabbed\tname': 'foo\tbar'},
		{'spaced name': 'newline', 'tabbed\tname': 'a brand new\nline'},
		{'spaced name': 'doublequote start', 'tabbed\tname': '"foo'},
		{'spaced name': 'doublequote inside', 'tabbed\tname': 'f"oo'},
		{'spaced name': 'doublequote end', 'tabbed\tname': 'foo"'},
		{'spaced name': 'singlequote start', 'tabbed\tname': "'foo"},
		{'spaced name': 'singlequote inside', 'tabbed\tname': "f'oo"},
		{'spaced name': 'singlequote end', 'tabbed\tname': "foo'"},
	])

	# test that escaping of ds name works in the right places
	comma_ds = mk_ds(',', ['ascii'], ['foo'])
	commaquote_ds = mk_ds(',"', ['ascii'], ['bar'])
	newline_ds = mk_ds('\n', ['ascii'], ['really?'])
	grep_text(['-t', ',', '-D', '', comma_ds], [['"' + comma_ds.replace('"', '""') + '"', 'foo']], sep=',')
	grep_text(['-t', ',', '-D', '', commaquote_ds], [['"' + commaquote_ds.replace('"', '""') + '"', 'bar']], sep=',')
	grep_text(['-t', ',', '-D', '', newline_ds], [[newline_ds.replace('\n', '\\n'), 'really?']], sep=',')
	# and the same thing but -f raw, so it's not escaped
	grep_text(['-t', ',', '-D', '-f', 'raw', '', comma_ds], [[comma_ds, 'foo']], sep=',')
	grep_text(['-t', ',', '-D', '-f', 'raw', '', commaquote_ds], [[commaquote_ds, 'bar']], sep=',')
	# this ends up looking like two lines
	grep_text(['-t', ',', '-D', '-f', 'raw', '', newline_ds], [[newline_ds.replace('\n', '')], [',really?']], sep=',')

	alltypes = mk_ds('alltypes',
		[
			'ascii',
			'bits32',
			'bits64',
			'bool',
			'bytes',
			'complex32',
			'complex64',
			'date',
			'datetime',
			'float32',
			'float64',
			'json',
			'number',
			'time',
			'unicode',
		], [
			'foo',
			11111,
			99999,
			True,
			b'\xff\x00octets',
			1+2j,
			1.5-0.5j,
			datetime.date(2021, 9, 20),
			datetime.datetime(2021, 9, 20, 1, 2, 3),
			0.125,
			1e42,
			[1, 2, 3, {'FOO': 'BAR'}, None],
			-2,
			datetime.time(4, 5, 6),
			'codepoints\x00\xe4',
		], [
			'',
			0,
			0,
			False,
			b'',
			0j,
			0j,
			datetime.date(1, 1, 1),
			datetime.datetime(1, 1, 1, 1, 1, 1),
			0.0,
			0.0,
			'json',
			0,
			datetime.time(1, 1, 1),
			'',
		],
	)
	grep_text(['json', alltypes], [['', 0, 0, 'False', '', '0j', '0j', '0001-01-01', '0001-01-01 01:01:01', '0.0', '0.0', 'json', 0, '01:01:01', '']])
	grep_text(['-g', 'json', 'foo', alltypes], [])
	grep_text(['-g', 'bytes', 'tet', alltypes, 'ascii', 'unicode'], [['foo', 'codepoints\x00\xe4']])
	grep_text(['-g', 'bytes', '\\x00', alltypes, 'bool'], [['True']])
	if PY3:
		# python2 doesn't really handle non-utf8 bytes
		grep_text(['-g', 'bytes', '\\udcff', alltypes, 'bool'], [['True']])
	grep_text(['--format=raw', '-g', 'json', '-i', 'foo', alltypes], [[b'foo', b'11111', b'99999', b'True', b'\xff\x00octets' if PY3 else b'\xef\xbf\xbd\x00octets', b'(1+2j)', b'(1.5-0.5j)', b'2021-09-20', b'2021-09-20 01:02:03', b'0.125', b'1e+42', b"[1, 2, 3, {'FOO': 'BAR'}, None]" if PY3 else b"[1, 2, 3, {u'FOO': u'BAR'}, None]", b'-2', b'04:05:06', b'codepoints\x00\xc3\xa4']], sep=b'\t', encoding=None)
	grep_json([':05:', alltypes, 'bool', 'time', 'unicode', 'bytes'], [{'bool': True, 'time': '04:05:06', 'unicode': 'codepoints\x00\xe4', 'bytes': '\udcff\x00octets' if PY3 else '\ufffd\x00octets'}])

	columns = [
		'ascii',
		'bits32',
		'bits64',
		'bytes',
		'complex32',
		'complex64',
		'date',
		'datetime',
		'float32',
		'float64',
		'int32',
		'int64',
		'json',
		'number',
		'time',
		'unicode',
	]
	d = mk_ds('d', columns,
		['42', 0, 0, b'', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '', 0, datetime.time(1, 1, 1), '',],
		['', 42, 0, b'', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '', 0, datetime.time(1, 1, 1), '',],
		['', 0, 42, b'', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '', 0, datetime.time(1, 1, 1), '',],
		['', 0, 0, b'42', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '', 0, datetime.time(1, 1, 1), '',],
		['', 0, 0, b'', 42+0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '', 0, datetime.time(1, 1, 1), '',],
		['', 0, 0, b'', 0j, 42+0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '', 0, datetime.time(1, 1, 1), '',],
		['', 0, 0, b'', 0j, 0j, datetime.date(42, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '', 0, datetime.time(1, 1, 1), '',],
		['', 0, 0, b'', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(42, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '', 0, datetime.time(1, 1, 1), '',],
		['', 0, 0, b'', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 42.0, 0.0, 0, 0, '', 0, datetime.time(1, 1, 1), '',],
		['', 0, 0, b'', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 42.0, 0, 0, '', 0, datetime.time(1, 1, 1), '',],
		['', 0, 0, b'', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 42, 0, '', 0, datetime.time(1, 1, 1), '',],
		['', 0, 0, b'', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 42, '', 0, datetime.time(1, 1, 1), '',],
		['', 0, 0, b'', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '42', 0, datetime.time(1, 1, 1), '',],
		['', 0, 0, b'', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '', 42, datetime.time(1, 1, 1), '',],
		['a', 0, 0, b'b', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '', 0, datetime.time(1, 1, 42), '',],
		['B', 0, 0, b'A', 0j, 0j, datetime.date(1, 1, 1), datetime.datetime(1, 1, 1, 1, 1, 1), 0.0, 0.0, 0, 0, '', 0, datetime.time(1, 1, 1), '42',],
	)
	want = [
		['42', 0, 0, '', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 0, '', 0, '01:01:01', ''],
		['', 42, 0, '', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 0, '', 0, '01:01:01', ''],
		['', 0, 42, '', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 0, '', 0, '01:01:01', ''],
		['', 0, 0, '42', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 0, '', 0, '01:01:01', ''],
		['', 0, 0, '', 42+0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 0, '', 0, '01:01:01', ''],
		['', 0, 0, '', 0j, 42+0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 0, '', 0, '01:01:01', ''],
		['', 0, 0, '', 0j, 0j, '0042-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 0, '', 0, '01:01:01', ''],
		['', 0, 0, '', 0j, 0j, '0001-01-01', '0042-01-01 01:01:01', 0.0, 0.0, 0, 0, '', 0, '01:01:01', ''],
		['', 0, 0, '', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 42.0, 0.0, 0, 0, '', 0, '01:01:01', ''],
		['', 0, 0, '', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 42.0, 0, 0, '', 0, '01:01:01', ''],
		['', 0, 0, '', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 42, 0, '', 0, '01:01:01', ''],
		['', 0, 0, '', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 42, '', 0, '01:01:01', ''],
		['', 0, 0, '', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 0, '42', 0, '01:01:01', ''],
		['', 0, 0, '', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 0, '', 42, '01:01:01', ''],
		['a', 0, 0, 'b', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 0, '', 0, '01:01:42', ''],
		['B', 0, 0, 'A', 0j, 0j, '0001-01-01', '0001-01-01 01:01:01', 0.0, 0.0, 0, 0, '', 0, '01:01:01', '42'],
	]
	grep_text(['42', d], want)
	def json_fixup(line):
		line = list(line)
		for key in ('complex32', 'complex64',):
			value = line[columns.index(key)]
			line[columns.index(key)] = [value.real, value.imag]
		return line
	want_json = [dict(zip(columns, json_fixup(line))) for line in want]
	grep_json(['42', d], want_json)
	grep_json(['-D', '42', d], [{'dataset': d, 'data': data} for data in want_json])
	grep_text(['-i', 'a', d], [
		['a', 0, 0, 'b', '0j', '0j', '0001-01-01', '0001-01-01 01:01:01', '0.0', '0.0', 0, 0, '', 0, '01:01:42', ''],
		['B', 0, 0, 'A', '0j', '0j', '0001-01-01', '0001-01-01 01:01:01', '0.0', '0.0', 0, 0, '', 0, '01:01:01', '42'],
	])
	grep_text(['-i', 'a', d, 'unicode', 'ascii'], [['', 'a']])
	grep_text(['-i', '-g', 'bytes', 'a', d, 'unicode', 'ascii'], [['42', 'B']])
	grep_json(['-g', 'bits32', '-g', 'ascii', '-D', '-L', '-S', '42', d], [
		{'dataset': d, 'sliceno': 0, 'lineno': 0, 'data': want_json[0]},
		{'dataset': d, 'sliceno': 1, 'lineno': 0, 'data': want_json[1]},
	])
	all_types = {n for n in _convfuncs if not n.startswith('parsed:')}
	if PY2:
		all_types.remove('pickle')
	assert used_types == all_types, 'Missing/extra column types: %r %r' % (all_types - used_types, used_types - all_types,)

	# test the smart tab mode with various lengths

	# with traditional tabs specified differently:
	# t       t       t       t       t       t       t       t       t
	# date    datetime        float32 float64 int32   int64
	# 0042-01-01      0042-01-01 01:01:01     0.0     0.0     0       0
	grep_text(['--color=always', '--tab-length=8/8/1', '-H', '42-01', d, 'date', 'datetime', 'float32', 'float64', 'int32', 'int64'], [
			hdrsepframe('date', '    ', 'datetime', '        ', 'float32', ' ', 'float64', ' ', 'int32', '   ', 'int64'),
			sepframe('00\x1b[31m42-01\x1b[39m-01', '      ', '0001-01-01 01:01:01', '     ', '0.0', '     ', '0.0', '     ', '0', '       ', '0'),
			sepframe('0001-01-01', '      ', '00\x1b[31m42-01\x1b[39m-01 01:01:01', '     ', '0.0', '     ', '0.0', '     ', '0', '       ', '0'),
		], sep='',
	)

	# same, but with min_len=2
	# with traditional tabs specified differently:
	# t       t       t       t       t       t       t       t       t
	# date    datetime        float32         float64         int32   int64
	# 0042-01-01      0042-01-01 01:01:01     0.0     0.0     0       0
	grep_text(['--color=always', '--tab-length=8/8/2', '-H', '42-01', d, 'date', 'datetime', 'float32', 'float64', 'int32', 'int64'], [
			hdrsepframe('date', '    ', 'datetime', '        ', 'float32', '         ', 'float64', '         ', 'int32', '   ', 'int64'),
			sepframe('00\x1b[31m42-01\x1b[39m-01', '      ', '0001-01-01 01:01:01', '     ', '0.0', '     ', '0.0', '     ', '0', '       ', '0'),
			sepframe('0001-01-01', '      ', '00\x1b[31m42-01\x1b[39m-01 01:01:01', '     ', '0.0', '     ', '0.0', '     ', '0', '       ', '0'),
		], sep='',
	)

	# with the default smart values
	# f               f               f               f               f               f
	# t       t       t       t       t       t       t       t       t       t       t
	# date            datetime        float32         float64         int32           int64
	# 0042-01-01      0042-01-01 01:01:01     0.0     0.0             0               0
	grep_text(['--color=always', '--tab-length=/', '-H', '42-01', d, 'date', 'datetime', 'float32', 'float64', 'int32', 'int64'], [
			hdrsepframe('date', '            ', 'datetime', '        ', 'float32', '         ', 'float64', '         ', 'int32', '           ', 'int64'),
			sepframe('00\x1b[31m42-01\x1b[39m-01', '      ', '0001-01-01 01:01:01', '     ', '0.0', '     ', '0.0', '             ', '0', '               ', '0'),
			sepframe('0001-01-01', '      ', '00\x1b[31m42-01\x1b[39m-01 01:01:01', '     ', '0.0', '     ', '0.0', '             ', '0', '               ', '0'),
		], sep='',
	)
	# same, but explicitly specified
	grep_text(['--color=always', '--tab-length=8/16/2', '-H', '42-01', d, 'date', 'datetime', 'float32', 'float64', 'int32', 'int64'], [
			hdrsepframe('date', '            ', 'datetime', '        ', 'float32', '         ', 'float64', '         ', 'int32', '           ', 'int64'),
			sepframe('00\x1b[31m42-01\x1b[39m-01', '      ', '0001-01-01 01:01:01', '     ', '0.0', '     ', '0.0', '             ', '0', '               ', '0'),
			sepframe('0001-01-01', '      ', '00\x1b[31m42-01\x1b[39m-01 01:01:01', '     ', '0.0', '     ', '0.0', '             ', '0', '               ', '0'),
		], sep='',
	)

	# with field_len=11 which is not a multiple of tab_len=8
	# f          f          f          f          f          f          f
	# t       t       t       t       t       t       t       t       t       t
	# date            datetime        float32 float64 int32   int64           bits32
	# 0042-01-01      0042-01-01 01:01:01     0.0     0.0     0       0       0
	grep_text(['--color=always', '--tab-length=8/11/1', '-H', '42-01', d, 'date', 'datetime', 'float32', 'float64', 'int32', 'int64', 'bits32'], [
			hdrsepframe('date', '            ', 'datetime', '        ', 'float32', ' ', 'float64', ' ', 'int32', '   ', 'int64', '           ', 'bits32'),
			sepframe('00\x1b[31m42-01\x1b[39m-01', '      ', '0001-01-01 01:01:01', '     ', '0.0', '     ', '0.0', '     ', '0', '       ', '0', '       ', '0'),
			sepframe('0001-01-01', '      ', '00\x1b[31m42-01\x1b[39m-01 01:01:01', '     ', '0.0', '     ', '0.0', '     ', '0', '       ', '0', '       ', '0'),
		], sep='',
	)

	# with field_len=13 which is not a multiple of tab_len=8
	# f            f            f            f            f            f            f
	# t       t       t       t       t       t       t       t       t       t       t
	# date            datetime        float32 float64         int32           int64   bits32
	# 0042-01-01      0042-01-01 01:01:01     0.0     0.0     0               0       0
	grep_text(['--color=always', '--tab-length=8/13/1', '-H', '42-01', d, 'date', 'datetime', 'float32', 'float64', 'int32', 'int64', 'bits32'], [
			hdrsepframe('date', '            ', 'datetime', '        ', 'float32', ' ', 'float64', '         ', 'int32', '           ', 'int64', '   ', 'bits32'),
			sepframe('00\x1b[31m42-01\x1b[39m-01', '      ', '0001-01-01 01:01:01', '     ', '0.0', '     ', '0.0', '     ', '0', '               ', '0', '       ', '0'),
			sepframe('0001-01-01', '      ', '00\x1b[31m42-01\x1b[39m-01 01:01:01', '     ', '0.0', '     ', '0.0', '     ', '0', '               ', '0', '       ', '0'),
		], sep='',
	)

	# exercise the parser a bit more (this is 4/8/3, written stupidly)
	# f       f       f       f       f       f       f       f       f
	# t   t   t   t   t   t   t   t   t   t   t   t   t   t   t   t   t
	# date    datetime    float32     float64     int32   int64   bits32
	# 0042-01-01      0001-01-01 01:01:01     0.0     0.0     0   0   0
	grep_text(['--color=always', '--tab-length=1/min_LENGTH=3/8', '-T', 'tablen=4', '-H', '42-01', d, 'date', 'datetime', 'float32', 'float64', 'int32', 'int64', 'bits32'], [
			hdrsepframe('date', '    ', 'datetime', '    ', 'float32', '     ', 'float64', '     ', 'int32', '   ', 'int64', '   ', 'bits32'),
			sepframe('00\x1b[31m42-01\x1b[39m-01', '      ', '0001-01-01 01:01:01', '     ', '0.0', '     ', '0.0', '     ', '0', '   ', '0', '   ', '0'),
			sepframe('0001-01-01', '      ', '00\x1b[31m42-01\x1b[39m-01 01:01:01', '     ', '0.0', '     ', '0.0', '     ', '0', '   ', '0', '   ', '0'),
		], sep='',
	)



description = 'return "Hello, World!"'


def synthesis():
	# The synthesis-function is run once. The return value is easy to
	# load into another job or build script without the need to
	# specify a filename.  (So we use the job that created the data as
	# reference, never again an arbitrary filename.)
	return 'Hello, World!'



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify that subjobs are allowed to nest (exactly) five levels.
'''

from accelerator import subjobs
from accelerator.dispatch import JobError

options = {'level': 0}

def synthesis():
	assert options.level < 5, "Too deep subjob nesting allowed"
	try:
		subjobs.build('test_subjobs_nesting', options={'level': options.level + 1})
	except JobError:
		assert options.level == 4, "Not enough subjob nesting allowed"



############################################################################
#                                                                          #
# Copyright (c) 2019-2020 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test dataset_checksum[_chain].
'''

from accelerator.dataset import DatasetWriter
from accelerator import subjobs
from accelerator import blob
from accelerator.compat import PY3

test_data = [
	("a", b"A", b"0", 0.42, 18, [1, 2, 3], u"a", u"A"),
	("b", b"B", b"1", 0.18, 42, "1, 2, 3", u"b", u"B"),
	# And one with difficult values.
	("c", b"\xe4", None, float('NaN'), -1, {1: 2, "3": 4, "foo": "bar", "bar": "|"}, u"\xe4", None),
]

def prepare():
	columns = dict(
		ascii="ascii",
		bytes="bytes",
		bytes_none=("bytes", True),
		float="float64",
		int="int64",
		json="json",
		unicode="unicode",
		unicode_none=("unicode", True),
	)
	if PY3:
		# z so it sorts last
		columns['zpickle'] = 'pickle'
		for ix, v in enumerate(test_data):
			test_data[ix] = v + ([ix, 'line %d' % (ix,), {'line': ix}, 42],)
		test_data[-1][-1][-1] = float('-inf')
	a = DatasetWriter(name="a", columns=columns)
	b = DatasetWriter(name="b", columns=columns, previous=a)
	c = DatasetWriter(name="c", columns=columns)
	return a, b, c, test_data

def analysis(sliceno, prepare_res):
	a, b, c, test_data = prepare_res
	if sliceno == 0:
		for data in test_data[:2]:
			a.write_list(data)
		for data in test_data[:1]:
			b.write_list(data)
		for data in test_data[::-1]:
			c.write_list(data)
	if sliceno == 1:
		for data in test_data[2:]:
			a.write_list(data)
		for data in test_data[-1:0:-1]:
			b.write_list(data)
		for data in test_data:
			c.write_list(data)

def ck(jid, method="dataset_checksum", **kw):
	jid = subjobs.build(method, datasets=dict(source=jid), options=kw)
	return blob.load(jobid=jid).sum

def synthesis(prepare_res):
	a, b, c, _ = prepare_res
	a = a.finish()
	b = b.finish()
	c = c.finish()
	a_sum = ck(a)
	b_sum = ck(b)
	c_sum = ck(c)
	assert a_sum == b_sum # only order differs, so sorted they are the same.
	assert a_sum != c_sum # each line twice in c
	ab_sum = ck(b, "dataset_checksum_chain")
	assert ab_sum == 0 # a_sum and b_sum are the same, so a_sum ^ b_sum == 0
	assert ab_sum != a_sum
	assert ab_sum != c_sum # chains and lines are handled differently.
	cc_sum = ck(c, "dataset_checksum_chain")
	assert cc_sum == c_sum # but a chain of one should be equal to that one.
	a_uns_sum = ck(a, sort=False)
	b_uns_sum = ck(b, sort=False)
	assert a_uns_sum != b_uns_sum # they are not the same order
	if PY3:
		# Check that the pickle column really was included and works.
		a_p_sum = ck(a, columns={'zpickle'})
		b_p_sum = ck(b, columns={'zpickle'})
		assert a_p_sum == b_p_sum # same values
		a_uns_p_sum = ck(a, columns={'zpickle'}, sort=False)
		b_uns_p_sum = ck(b, columns={'zpickle'}, sort=False)
		assert a_uns_p_sum != b_uns_p_sum # but they are not the same order



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2021 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

from accelerator.compat import PY3, unquote_plus
from accelerator.compat import urlopen, Request, URLError, HTTPError
from accelerator.extras import json_encode, json_decode
from accelerator.error import ServerError, UrdError, UrdPermissionError, UrdConflictError
from accelerator import g, __version__ as ax_version

if PY3:
	from urllib.request import install_opener, build_opener, AbstractHTTPHandler
	from http.client import HTTPConnection
else:
	from urllib2 import install_opener, build_opener, AbstractHTTPHandler
	from httplib import HTTPConnection

import sys
import time
import socket

class UnixHTTPConnection(HTTPConnection):
	def __init__(self, host, *a, **kw):
		HTTPConnection.__init__(self, 'localhost', *a, **kw)
		self.unix_path = unquote_plus(host.split(':', 1)[0])

	def connect(self):
		s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
		s.connect(self.unix_path)
		self.sock = s

class UnixHTTPHandler(AbstractHTTPHandler):
	def unixhttp_open(self, req):
		return self.do_open(UnixHTTPConnection, req)

	unixhttp_request = AbstractHTTPHandler.do_request_

install_opener(build_opener(UnixHTTPHandler))


import bottle

# The standard bottle WaitressServer can't handle unix sockets and doesn't set threads.
class WaitressServer(bottle.ServerAdapter):
	def run(self, handler):
		from waitress import create_server
		if self.port:
			kw = dict(host=self.host, port=self.port)
		else:
			kw = dict(unix_socket=self.host, unix_socket_perms='777')
		server = create_server(handler, threads=12, **kw)
		server.run()


def call(url, data=None, fmt=json_decode, headers={}, server_name='server', retries=4, quiet=False):
	if data is not None and not isinstance(data, bytes):
		data = json_encode(data)
	err = None
	req = Request(url, data=data, headers=headers)
	for attempt in range(1, retries + 2):
		resp = None
		try:
			r = urlopen(req)
			try:
				resp = r.read()
				if server_name == 'server' and g.running in ('build', 'shell',):
					s_version = r.headers['Accelerator-Version'] or '<unknown (old)>'
					if s_version != ax_version:
						# Nothing is supposed to catch this, so just print and die.
						print('Server is running version %s but we are running version %s' % (s_version, ax_version,), file=sys.stderr)
						exit(1)
				if PY3:
					resp = resp.decode('utf-8')
				# It is inconsistent if we get HTTPError or not.
				# It seems we do when using TCP sockets, but not when using unix sockets.
				if r.getcode() >= 400:
					raise HTTPError(url, r.getcode(), resp, {}, None)
				return fmt(resp)
			finally:
				try:
					r.close()
				except Exception:
					pass
		except HTTPError as e:
			if resp is None and e.fp:
				resp = e.fp.read()
				if PY3:
					resp = resp.decode('utf-8')
			msg = '%s says %d: %s' % (server_name, e.code, resp,)
			if server_name == 'urd' and 400 <= e.code < 500:
				if e.code == 401:
					err = UrdPermissionError()
				if e.code == 409:
					err = UrdConflictError()
				break
			if server_name == 'server' and e.code != 503 and resp:
				return fmt(resp)
		except URLError:
			# Don't say anything the first times, because the output
			# tests get messed up if this happens during them.
			if attempt < retries - 1:
				msg = None
			else:
				msg = 'error contacting ' + server_name
		except ValueError as e:
			msg = 'Bad data from %s, %s: %s' % (server_name, type(e).__name__, e,)
		if msg and not quiet:
			print(msg, file=sys.stderr)
		if attempt < retries + 1:
			time.sleep(attempt / 15)
			if msg and not quiet:
				print('Retrying (%d/%d).' % (attempt, retries,), file=sys.stderr)
	else:
		if not quiet:
			print('Giving up.', file=sys.stderr)
	if err:
		raise err
	if server_name == 'urd':
		raise UrdError(msg)
	else:
		raise ServerError(msg)



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = r'''
Collects values for dataset_fanout.

You probably don't want to call this yourself.
'''

from accelerator import OptionString
from accelerator.compat import unicode, imap


options = {
	'column': OptionString,
	'length': -1,
}

datasets = ('source',)

jobs = ('previous',)


def analysis(sliceno):
	chain = datasets.source.chain(stop_ds={jobs.previous: 'source'}, length=options.length)
	return set(imap(unicode, chain.iterate(sliceno, options.column)))

def synthesis(analysis_res):
	return analysis_res.merge_auto()



############################################################################
#                                                                          #
# Copyright (c) 2022 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = r'''
Test DatasetWriter.add(..., default=...)
'''

from datetime import datetime, timedelta

def synthesis(job):
	now = datetime.now()
	later = now + timedelta(days=1, hours=1)
	for   t         , good_value    , default_value, bad1        , bad2 in (
		('int64'    , 1             , 2            , float('inf'), None),
		('int32'    , 1             , 2            , float('inf'), None),
		('bits64'   , 1             , 2            , -1          , None),
		('bits32'   , 1             , 2            , -1          , None),
		('float64'  , 0.1           , 0.2          , 'bad'       , None),
		('float32'  , 1             , 2            , 'bad'       , None),
		('number'   , 1             , 2.1          , 'bad'       , None),
		('complex64', 0.1+1j        , 2-0.2j       , 'bad'       , None),
		('complex32', 1+1j          , 2-2j         , 'bad'       , None),
		('bool'     , False         , True         , 2           , None),
		('datetime' , now           , later        , 95          , None),
		('date'     , now.date()    , later.date() , ()          , None),
		('time'     , now.time()    , later.time() , ''          , None),
		('json'     , {'foo': 'bar'}, [1, 2, 3]    , datetime    , now),
		('ascii'    , 'foo'         , 'bar'        , u'\xe4'     , b'\xe4'),
		('bytes'    , b'foo'        , b'bar'       , u'baz'      , None),
		('unicode'  , u'\xe4'       , u'\xe5'      , b'baz'      , None),
	):
		for bad_default in (bad1, bad2):
			dw = job.datasetwriter(name='failme')
			try:
				dw.add('data', t, default=bad_default)
				dw.get_split_write()
				raise Exception('%s accepted %r as default value' % (t, bad_default,))
			except (TypeError, ValueError, OverflowError):
				pass
			dw.discard()
		dw = job.datasetwriter(name=t, allow_missing_slices=True)
		dw.add('data', t, default=default_value)
		dw.set_slice(0)
		dw.write(good_value)
		dw.write(bad1)
		dw.write(bad2)
		ds = dw.finish()
		want = [good_value, default_value, default_value]
		got = list(ds.iterate(0, 'data'))
		assert got == want, '%s failed, wanted %r but got %r' % (ds.quoted, want, got,)
		if not t.startswith('bits'):
			# no none_support in the bits types
			dw = job.datasetwriter(name=t + ' default=None', allow_missing_slices=True)
			dw.add('data', t, default=None, none_support=True)
			dw.set_slice(0)
			dw.write(good_value)
			dw.write(bad1)
			dw.write(bad2)
			ds = dw.finish()
			want = [good_value, None, None]
			got = list(ds.iterate(0, 'data'))
			assert got == want, '%s failed, wanted %r but got %r' % (ds.quoted, want, got,)

			# make sure default=None hashes correctly
			if t != 'json':
				dw = job.datasetwriter(name=t + ' default=None hashed', hashlabel='data')
				dw.add('data', t, default=None, none_support=True)
				w = dw.get_split_write()
				w(bad1)
				w(None)
				w(bad1) # bad2 might be None, so don't use that.
				ds = dw.finish()
				want = [None, None, None]
				got = list(ds.iterate(0, 'data'))
				assert got == want, '%s slice 0 failed, wanted %r but got %r' % (ds.quoted, want, got,)



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify filename (sliced and unsliced) and gzip in csvexport.
'''

from accelerator import subjobs
import gzip

def synthesis(job):
	dw = job.datasetwriter(name='a')
	dw.add('a', 'int32')
	w = dw.get_split_write()
	w(0)
	w(1)
	w(2)
	a = dw.finish()

	for filename, sliced, out_filename, open_func in (
		('name', False, 'name', open),
		('name', True, 'name.1', open),
		('name%dend', True, 'name1end', open),
		('name.gz', False, 'name.gz', gzip.open),
		('name.gz', True, 'name.gz.1', gzip.open),
		('a%02d.gz.b', True, 'a01.gz.b', gzip.open),
		('a%02d.gz.b', False, 'a%02d.gz.b', gzip.open),
		('name.gzonk', False, 'name.gzonk', open),
	):
		job = subjobs.build('csvexport', filename=filename, sliced=sliced, source=a, labels=['a'])
		fn = job.filename(out_filename)
		with open_func(fn, mode='rb') as fh:
			got = fh.read()
		if sliced:
			want = b'a\n1\n'
		else:
			want = b'a\n0\n1\n2\n'
		assert want == got, 'wanted %r, got %r in %s' % (want, got, fn)



from datetime import date
from pprint import pprint

options = dict(
	day1=date(1970, 1, 1),         # reqires a date (or None). Here, a default value is set.
	day2=date,                     # requires a date, will be None if not assigned
	something=None,                # accepts anything (that works in json)
	text='',                       # requires a string (or None).  Here, the empty string is default.
	number=int,                    # requires an int (or None)
	seq=[],                        # reqires a list
	cplx=dict(str=dict(str=str)),  # reqires a dict of string to dict of string to string
)


def synthesis():
	pprint(options)



from os.path import dirname, join
from .printer import prt

description = "Tutorial: Import a set of files and persistently store all references and dependencies."


filenames = ('data.csv', 'data2.csv', 'data3.csv')
# Again, files are stored in same directory as this python file.
# In a real project, set "input directory" in config file instead!
filenames = (join(dirname(__file__), x) for x in filenames)


def main(urd):
	prt.source(__file__)
	prt()
	prt('Import a list of files and store references in the persistent')
	prt('Urd database for later use.')
	prt()
	key = 'example_tutorial05'
	urd.truncate(key, 0)
	for ts, filename in enumerate(filenames, 1):
		prt.plain(filename, ts)
		urd.begin(key, ts, caption=filename)
		latest = urd.latest(key).joblist
		imp = urd.build(
			'csvimport',
			filename=filename,
			previous=latest.get('csvimport'),
		)
		imp = urd.build(
			'dataset_type',
			source=imp,
			previous=latest.get('dataset_type'),
			column2type=dict(
				Date='date:%Y-%m-%d',
				String='unicode:utf-8',
				Int='number',
				Float='float64',
			),
		)
		imp = urd.build(
			'dataset_sort',
			source=imp,
			sort_columns='Date',
			previous=latest.get('dataset_sort'),
		)
		imp = urd.build(
			'dataset_hashpart',
			source=imp,
			hashlabel='String',
			previous=latest.get('dataset_hashpart'),
		)
		urd.finish(key)

	prt()
	prt('''
		Now, references to everything that has been built, including
		job dependencies, is stored in the Urd server using the.
		key "<user>/%s".
	''' % (key,))

	prt()
	with prt.header('View all Urd lists.'):
		prt('To view all Urd "lists", type')
		prt.command('ax urd')
		prt('Each key is composed by <user>/<listname>:')
		for item in urd.list():
			prt.output(item)

	prt()
	with prt.header('Inspecting all sessions in a list.'):
		prt('All sessions are timestamped.  To see all timestamps')
		prt('and captions in "%s" since timestamp zero, do' % (key,))
		prt.command('ax urd %s/since/0' % (key,))
		prt('or equivalently')
		prt.command('ax urd %s/' % (key,))
		prt('the output from this command looks something like this (try it)')
		for ts in urd.since(key, 0):
			prt.output(ts, urd.peek(key, ts).caption)

	prt()
	with prt.header('Inspecting individual Urd items.'):
		prt('We can look at an individual Urd-item like this')
		prt.command('ax urd %s/3' % (key,))
		prt('which corresponds to the list in key "%s" at timestamp 3.' % (key,))
		prt('(Timestamps can be dates, datetimes, integers, or tuples')
		prt('of date/datetimes and integers.)')

	prt()
	prt('''
		The data is imported into jobs in a workdir, and references
		to these jobs are recorded in the Urd persistent transaction-log
		database.
	''')



from os.path import dirname, join
from .printer import prt

description = "Tutorial: Iterate over a chain of datasets."


filenames = ('data.csv', 'data2.csv', 'data3.csv')
# Again, files are stored in same directory as this python file.
# In a real project, set "input directory" in config file instead!
filenames = (join(dirname(__file__), x) for x in filenames)


def main(urd):
	prt.source(__file__)
	prt()
	with prt.header('Iterate over a dataset chain.'):
		prt('''
			We've copied the import loop from "build_tutorial03.py" in
			order to get a reference to the chained dataset we've created.

			(Soon, we'll show how to do this in a much more elegant way!)
		''')
		prt()

	for filename in filenames:
		prt.plain(filename)
		imp = urd.build(
			'csvimport',
			filename=filename,
			previous=urd.joblist.get('csvimport'),
		)
		imp = urd.build(
			'dataset_type',
			source=imp,
			previous=urd.joblist.get('dataset_type'),
			column2type=dict(
				Date='date:%Y-%m-%d',
				String='unicode:utf-8',
				Int='number',
				Float='float64',
			),
		)
		imp = urd.build(
			'dataset_sort',
			source=imp,
			sort_columns='Date',
			previous=urd.joblist.get('dataset_sort'),
		)
		imp = urd.build(
			'dataset_hashpart',
			source=imp,
			hashlabel='String',
			previous=urd.joblist.get('dataset_hashpart'),
		)

	with prt():
		prt('''
			The variable "imp" now hold a reference to the last built method,
			which happens to be dataset_hashpart of the data from "data3.csv".
			Because of the chaining, all imported data is available through
			this reference.
		''')

		prt()
		prt('Now, let\'s iterate over all imported data')
		prt()
		job = urd.build('dsexample_iteratechain', source=imp)
		prt()
		prt('...and the result is (something printed by the job to stdout)')
		prt.output(job.output())
		prt()
		prt('Try')
		prt.command('ax job -O', job)
		prt('to print this output from the %s job to the terminal.' % (job.method,))



############################################################################
#                                                                          #
# Copyright (c) 2021-2022 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import sys, os
from functools import partial

from accelerator.compat import PY2
from accelerator.error import ColourError

# all gray colours in the 256 colour palette in intensity order
_gray2idx = tuple(ix for _, ix in sorted(
	[(0, 16)] +
	[(gray * 10 + 8, gray + 232) for gray in range(24)] +
	[(rgb * 40 + 55, rgb * 43 + 16) for rgb in (1, 2, 3, 4, 5)]
))

class Colour:
	"""Constants and functions for colouring output.

	Available as constants named .COLOUR, functions named .colour and
	as direct calls on the object taking (value, *attrs).
	Colours are BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE,
	These can be prefixed with BRIGHT and/or suffixed with BG.
	DEFAULT[BG] restores the default colour.

	BOLD, FAINT, ITALIC, UNDERLINE, BLINK, INVERT, and STRIKE are also
	available. These can be prefixed with NOT to turn them off.

	When using the constants, you should usually end with .RESET.

	>>> colour.RED + 'foo' + colour.DEFAULT == colour.red('foo') == colour('foo', 'red')

	colour(v, 'red', 'bold') and similar produce shorter sequences than other
	ways of combining several attributes.

	You can also use:
		colour(v, '#RGB[bg]') (0 - 5) ("256 colour" mode)
		colour(v, '#GG[bg]') (00 - 1D) (grayscale from "256 colour" mode)
		colour(v, 'XNN[bg]') (00 - FF) (directly specifying a "256 colour" index)
		colour(v, '#RRGGBB[bg]') (00 - FF) (but terminal support is not great)

	Finally you can use names you put in the config file, some of which
	have default values. This also has a fallback system, so 'foo/bar' will
	fall back to 'bar' if 'foo/bar' is not defined. The defaults can be
	found in shell/__init__.py (search for "configuration defaults").

	In the configuration file you can also use <pre >post to just put
	those literal strings around the value. e.g.
	[colour]
		grep/header = <\e[4:3m >\e[24m
	if your terminal supports such extended codes.

	The functions take force=True to return escape sequences even if colour
	is disabled and reset=True to reset all attributes before (and after)
	this sequence. (By default only the changed attributes are reset.)
	"""

	def __init__(self):
		self._names = {}
		self._all = dict(
			BOLD='1',
			FAINT='2',
			ITALIC='3',
			UNDERLINE='4',
			BLINK='5',
			INVERT='7',
			STRIKE='9',
		)
		self._all.update([('NOT' + k, '2' + v) for k, v in self._all.items()])
		self._all['NOTBOLD'] = '22'
		for num, name in enumerate([
			'BLACK', 'RED', 'GREEN', 'YELLOW',
			'BLUE', 'MAGENTA', 'CYAN', 'WHITE',
		]):
			for prefix, base in (('', 30), ('BRIGHT', 90)):
				self._all[prefix + name] = str(base + num)
				self._all[prefix + name + 'BG'] = str(base + 10 + num)
		self._all['DEFAULT'] = '39'
		self._all['DEFAULTBG'] = '49'
		for k in self._all:
			setattr(self, k.lower(), partial(self._single, k))
		self._on = {k: '\x1b[%sm' % (v,) for k, v in self._all.items()}
		self._on['RESET'] = '\x1b[m'
		self._all['RESET'] = ''
		if PY2:
			self._on = {k.encode('ascii'): v.encode('ascii') for k, v in self._on.items()}
			self._off = dict.fromkeys(self._on, b'')
		else:
			self._off = dict.fromkeys(self._on, '')
		self.enable()
		self.__all__ = [k for k in dir(self) if not k.startswith('_')]

	def configure_from_environ(self, environ=None, stdout=None):
		# trying to support both https://no-color.org/ and https://bixense.com/clicolors/
		if environ is None:
			environ = os.environ
		if environ.get('CLICOLOR_FORCE', '0') != '0':
			self.enable()
		elif environ.get('NO_COLOR') is not None:
			self.disable()
		elif environ.get('CLICOLOR', '1') != '0' and (stdout or sys.stdout).isatty():
			self.enable()
		else:
			self.disable()

	def enable(self):
		"Turn colours on"
		self.__dict__.update(self._on)
		self.enabled = True

	def disable(self):
		"Turn colours off (make all constants empty)"
		self.__dict__.update(self._off)
		self.enabled = False

	def _single(self, attr, value, reset=False, force=False):
		return self(value, attr, reset=reset, force=force)

	def _expand_names(self, things):
		for thing in things:
			orig_thing = thing
			if thing not in self._names and '/' in thing:
				thing = thing.split('/', 1)[1]
			if thing in self._names:
				for a in self._names[thing]:
					yield orig_thing, a
			else:
				yield orig_thing, thing

	def _literal_split(self, pieces):
		have = []
		for piece in pieces:
			if piece.startswith('<'):
				if have:
					yield '\x1b[' + ';'.join(have) + 'm'
					have = []
				yield piece[1:]
			else:
				have.append(piece)
		if have:
			yield '\x1b[' + ';'.join(have) + 'm'

	# When we drop python 2 we can change this to use normal keywords
	def __call__(self, value, *attrs, **kw):
		bad_kw = set(kw) - {'force', 'reset'}
		if bad_kw:
			raise TypeError('Unknown keywords %r' % (bad_kw,))
		if not attrs:
			raise TypeError('specify at least one attr')
		if (self.enabled or kw.get('force')):
			if kw.get('reset'):
				pre = ['0']
			else:
				pre = []
			post = set()
			literal_post = ''
			a_it = self._expand_names(attrs)
			for a_src, a in a_it:
				if not a:
					raise ColourError('%r expanded to nothing' % (a_src,))
				if a.startswith('>'):
					raise ColourError('A >post needs a preceding <pre (expanded %r from %r)' % (a, a_src,))
				if a.startswith('<'):
					try:
						a_post_src, a_post = next(a_it)
					except StopIteration:
						a_post_src = a_post = ''
					if not a_post.startswith('>') or a_src != a_post_src:
						raise ColourError('A <pre needs a following >post (expanded %r from %r)' % (a, a_src,))
					literal_post += a_post[1:]
					pre.append(a)
					continue
				want = a.upper()
				default = self._all['DEFAULTBG' if want.endswith('BG') else 'DEFAULT']
				if want[0] in '#X':
					if want.endswith('BG'):
						prefix = '48'
						want = want[:-2]
					else:
						prefix = '38'
					try:
						if want[0] == '#' and len(want) == 7:
							r, g, b = (str(int(w, 16)) for w in (want[1:3], want[3:5], want[5:7]))
							part = (prefix, '2', r, g, b)
						else:
							if want[0] == '#' and len(want) == 4:
								r, g, b = (int(w, 16) for w in want[1:])
								assert 0 <= r <= 5 and 0 <= g <= 5 and 0 <= b <= 5
								idx = r * 36 + g * 6 + b + 16
							elif want[0] == '#' and len(want) in (2, 3):
								gg = int(want[1:], 16)
								assert 0 <= gg < len(_gray2idx)
								idx = _gray2idx[gg]
							elif want[0] == 'X' and len(want) in (2, 3):
								idx = int(want[1:], 16)
								assert 0 <= idx <= 255
							else:
								raise ValueError()
							part = (prefix, '5', str(idx))
					except (ValueError, AssertionError):
						raise ColourError('Bad colour spec %r (from %r)' % (a, a_src,))
					pre.append(':'.join(part))
					post.add(default)
				else:
					if want not in self._all:
						raise ColourError('Unknown colour/attr %r (from %r)' % (a, a_src,))
					pre.append(self._all[want])
					post.add(self._all.get('NOT' + want, default))
			pre = ''.join(self._literal_split(pre))
			if kw.get('reset'):
				post = '\x1b[m' + literal_post
			elif post:
				post = '\x1b[' + ';'.join(sorted(post)) + 'm' + literal_post
			else:
				post = literal_post
		else:
			pre = post = ''
		if isinstance(value, bytes):
			return b'%s%s%s' % (pre.encode('utf-8'), value, post.encode('utf-8'),)
		return '%s%s%s' % (pre, value, post,)

colour = Colour()
colour.configure_from_environ()



############################################################################
#                                                                          #
# Copyright (c) 2019-2021 Carl Drougge                                     #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
CSV file to dataset.

Read a CSV file (can be gziped), with any single iso-8859-1 character
separator (including \0) or no separator at all (always one field per line)
with or without quotes. Any single iso-8859-1 character or both \n and \r\n
(default) as newline. Labels from first line or specified in options.

If allow_bad is set also creates a "bad" dataset containing lineno and data
from bad lines.

If comment, skip_lines or skip_empty_lines is set also creates a "skipped"
dataset containing lineno and data from skipped lines.

If you want lineno for good lines too set lineno_label.

There is no support for multi-line quoted fields. (But if you control the
writing side try using something like \x1e or \0 instead of newline.)
'''


import os
from multiprocessing import Process
from threading import Thread
import struct
import locale

from accelerator import OptionString, DotDict
from accelerator.dsutil import typed_reader
from accelerator.compat import setproctitle, uni
from . import csvimport

depend_extra = (csvimport,)

options = dict(
	filename          = OptionString,
	separator         = ',',   # Single iso-8859-1 character or empty for a single field.
	comment           = '',    # Single iso-8859-1 character or empty, lines beginning with this character are ignored.
	newline           = '',    # Empty means \n or \r\n, or you can specify any single iso-8859-1 character.
	quotes            = '',    # Empty or False means no quotes, True means both ' and ", any other character means itself.
	labelsonfirstline = True,
	labels            = [],    # Mandatory if not labelsonfirstline, always sets labels if set.
	strip_labels      = False, # Do .strip() on all labels (happens before rename).
	rename            = {},    # Labels to replace (if they are in the file) (happens before discard).
	discard           = set(), # Labels to not include (if they are in the file)
	lineno_label      = "",    # Label of column to store line number in (not stored if empty).
	allow_bad         = False, # Still succeed if some lines have too few/many fields or bad quotes
	                           # creates a "bad" dataset containing lineno and data from the bad lines.
	allow_extra_empty = False, # Still consider a line good if it has extra empty fields at the end.
	skip_lines        = 0,     # skip this many lines at the start of the file.
	skip_empty_lines  = False, # ignore empty lines
	compression       = 6,     # gzip level
)

datasets = ('previous', )


cstuff = csvimport.init()

def reader_status(status_fd, update):
	# try to get nicer number formating
	try:
		locale.resetlocale(locale.LC_NUMERIC)
	except Exception:
		pass
	count = 0
	while True:
		update('{0:n} lines read'.format(count))
		data = os.read(status_fd, 8)
		if not data:
			break
		count = struct.unpack("=Q", data)[0]

def reader_process(slices, filename, write_fds, labels_fd, success_fd, status_fd, comment_char, lf_char):
	# Terrible hack - try to close FDs we didn't want in this process.
	# (This is important, if the main process dies this won't be
	# detected if we still have these open.)
	keep_fds = set(write_fds)
	keep_fds.add(labels_fd)
	keep_fds.add(success_fd)
	keep_fds.add(status_fd)
	# a few extra to be safe.
	for fd in range(3, max(keep_fds) + 32):
		if fd not in keep_fds:
			try:
				os.close(fd)
			except OSError:
				pass
	setproctitle("reader")
	os.dup2(success_fd, 2) # reader writes errors to stderr
	os.close(success_fd)
	success_fd = 2
	res = cstuff.backend.reader(filename.encode("utf-8"), slices, options.skip_lines, options.skip_empty_lines, write_fds, labels_fd, status_fd, comment_char, lf_char)
	if not res:
		os.write(success_fd, b"\0")
	os.close(success_fd)

def char2int(name, empty_value, specials="empty"):
	char = options.get(name)
	if not char:
		return empty_value
	msg = "%s must be a single iso-8859-1 character (or %s)" % (name, specials,)
	if isinstance(char, bytes):
		char = uni(char)
	try:
		char = char.encode("iso-8859-1")
	except UnicodeEncodeError:
		raise Exception(msg)
	assert len(char) == 1, msg
	return cstuff.backend.char2int(char)

def import_slice(fallback_msg, fd, sliceno, slices, field_count, out_fns, gzip_mode, separator, r_num, quote_char, lf_char, allow_bad, allow_extra_empty):
	fn = "import.success.%d" % (sliceno,)
	fh = open(fn, "wb+")
	real_stderr = os.dup(2)
	try:
		os.dup2(fh.fileno(), 2)
		res = cstuff.backend.import_slice(*cstuff.bytesargs(fd, sliceno, slices, field_count, out_fns, gzip_mode, separator, r_num, quote_char, lf_char, allow_bad, allow_extra_empty))
		os.dup2(real_stderr, 2)
		fh.seek(0)
		msg = fh.read().decode("utf-8", "replace")
		if msg or res:
			raise Exception(msg.strip() or fallback_msg)
	finally:
		os.dup2(real_stderr, 2)
		os.close(real_stderr)
		fh.close()
		os.unlink(fn)

def prepare(job, slices):
	# use 256 as a marker value, because that's not a possible char value (assuming 8 bit chars)
	lf_char = char2int("newline", 256)
	# separator uses lf_char or \n as the empty value, because memchr might mishandle 256.
	separator = char2int("separator", 10 if lf_char == 256 else lf_char)
	comment_char = char2int("comment", 256)
	if options.quotes == 'True':
		quote_char = 256
	elif options.quotes == 'False':
		quote_char = 257
	else:
		quote_char = char2int("quotes", 257, "True/False/empty")
	filename = os.path.join(job.input_directory, options.filename)
	orig_filename = filename
	assert 1 <= options.compression <= 9

	# To get a more useful error if the file doesn't exist or similar
	open(filename, 'rb').close()

	fds = [os.pipe() for _ in range(slices)]
	read_fds = [t[0] for t in fds]
	write_fds = [t[1] for t in fds]

	if options.labelsonfirstline:
		labels_rfd, labels_wfd = os.pipe()
	else:
		labels_wfd = -1
	success_fh = open("reader.success", "wb+")
	status_rfd, status_wfd = os.pipe()

	p = Process(target=reader_process, name="reader", args=(slices, filename, write_fds, labels_wfd, success_fh.fileno(), status_wfd, comment_char, lf_char))
	p.start()
	for fd in write_fds:
		os.close(fd)
	os.close(status_wfd)

	if options.labelsonfirstline:
		os.close(labels_wfd)
		# re-use import logic
		out_fns = ["labels"]
		r_num = cstuff.mk_uint64(3)
		try:
			import_slice("c backend failed in label parsing", labels_rfd, -1, -1, -1, out_fns, b"wb1", separator, r_num, quote_char, lf_char, 0, 0)
		finally:
			os.close(labels_rfd)
		if os.path.exists("labels"):
			with typed_reader("bytes")("labels") as fh:
				labels_from_file = [lab.decode("utf-8", "backslashreplace") for lab in fh]
			os.unlink("labels")
		else:
			labels_from_file = []
	else:
		labels_from_file = None

	labels = options.labels or labels_from_file
	if options.allow_extra_empty:
		while labels and labels[-1] == '':
			labels.pop()
	assert labels, "No labels"
	if options.strip_labels:
		labels = [x.strip() for x in labels]
	labels = [options.rename.get(x, x) for x in labels]
	assert len(labels) == len(set(labels)), "Duplicate labels: %r" % (labels,)

	dw = job.datasetwriter(
		columns={n: 'bytes' for n in labels if n not in options.discard},
		filename=orig_filename,
		caption='csvimport of ' + orig_filename,
		previous=datasets.previous,
		meta_only=True,
	)
	if options.lineno_label:
		dw.add(options.lineno_label, "int64")

	def dsprevious(name):
		if datasets.previous and datasets.previous.name == 'default':
			from accelerator.error import NoSuchDatasetError
			try:
				return datasets.previous.job.dataset(name)
			except NoSuchDatasetError:
				return None
		return None

	if options.allow_bad:
		bad_dw = job.datasetwriter(
			name="bad",
			filename=orig_filename,
			columns=dict(lineno="int64", data="bytes"),
			caption='bad lines from csvimport of ' + orig_filename,
			previous=dsprevious('bad'),
			meta_only=True,
		)
	else:
		bad_dw = None

	if options.comment or options.skip_lines or options.skip_empty_lines:
		skipped_dw = job.datasetwriter(
			name="skipped",
			filename=orig_filename,
			columns=dict(lineno="int64", data="bytes"),
			caption='skipped lines from csvimport of ' + orig_filename,
			previous=dsprevious('skipped'),
			meta_only=True,
		)
	else:
		skipped_dw = None

	return separator, quote_char, lf_char, filename, orig_filename, labels, dw, bad_dw, skipped_dw, read_fds, success_fh, status_rfd,

def analysis(sliceno, slices, prepare_res, update_top_status):
	separator, quote_char, lf_char, filename, _, labels, dw, bad_dw, skipped_dw, fds, _, status_fd, = prepare_res
	if sliceno == 0:
		t = Thread(
			target=reader_status,
			args=(status_fd, update_top_status),
			name='reader status',
		)
		t.daemon = True
		t.start()
	else:
		os.close(status_fd)
	# Close the FDs for all other slices.
	# Not techically necessary, but it feels like a good idea.
	for ix, fd in enumerate(fds):
		if ix != sliceno:
			os.close(fd)
	out_fns = []
	for label in labels:
		if label in options.discard:
			out_fns.append(cstuff.NULL)
		else:
			out_fns.append(dw.column_filename(label))
	for extra_dw in (bad_dw, skipped_dw):
		if extra_dw:
			for n in ("lineno", "data"):
				out_fns.append(extra_dw.column_filename(n))
		else:
			out_fns.append(cstuff.NULL)
			out_fns.append(cstuff.NULL)
	if options.lineno_label:
		out_fns.append(dw.column_filename(options.lineno_label))
	else:
		out_fns.append(cstuff.NULL)
	r_num = cstuff.mk_uint64(3) # [good_count, bad_count, comment_count]
	gzip_mode = b"wb%d" % (options.compression,)
	try:
		import_slice("c backend failed in slice %d" % (sliceno,), fds[sliceno], sliceno, slices, len(labels), out_fns, gzip_mode, separator, r_num, quote_char, lf_char, options.allow_bad, options.allow_extra_empty)
	finally:
		os.close(fds[sliceno])
	return list(r_num)

def synthesis(prepare_res, analysis_res):
	separator, _, _, filename, _, labels, dw, bad_dw, skipped_dw, fds, success_fh, _, = prepare_res
	# Analysis may have gotten a perfectly legitimate EOF if something
	# went wrong in the reader process, so we need to check that all
	# went well.
	reader_res = []
	try:
		success_fh.seek(0)
		reader_res = success_fh.read()
	except OSError:
		pass
	if reader_res != b"\0":
		reader_res = reader_res.decode("utf-8", "replace").strip("\r\n \t\0")
		raise Exception(reader_res or "Reader process failed")
	success_fh.close()
	os.unlink("reader.success")
	good_counts = []
	bad_counts = []
	skipped_counts = []
	for sliceno, (good_count, bad_count, skipped_count) in enumerate(analysis_res):
		dw.set_lines(sliceno, good_count)
		if bad_dw:
			bad_dw.set_lines(sliceno, bad_count)
		if skipped_dw:
			skipped_dw.set_lines(sliceno, skipped_count)
		good_counts.append(good_count)
		bad_counts.append(bad_count)
		skipped_counts.append(skipped_count)
	for dw in (bad_dw, skipped_dw, dw,):
		if dw:
			dw.set_compressions("gzip")
	return DotDict(
		num_lines=sum(good_counts),
		lines_per_slice=good_counts,
		num_broken_lines=sum(bad_counts),
		broken_lines_per_slice=bad_counts,
		num_skipped_lines=sum(skipped_counts),
		skipped_lines_per_slice=skipped_counts,
	)



############################################################################
#                                                                          #
# Copyright (c) 2019-2020 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test JobWithFile file loading.
Pickle and json, sliced and unsliced.
'''

from accelerator import JobWithFile, build, blob
import json

options = dict(
	file=JobWithFile,
	inner=False,
)

def analysis(sliceno, job):
	data = {sliceno}
	if options.inner:
		if options.file.sliced:
			value = options.file.load(sliceno)
			assert value == data
			value = blob.load(options.file.filename(sliceno))
			assert value == data
		else:
			try:
				options.file.load(sliceno)
				raise Exception("Allowed sliced load of unsliced file")
			except AssertionError:
				pass
		job.save({'inner': sliceno}, 'inner.pickle', sliceno, temp=False)
		job.json_save({'inner': sliceno}, 'inner.json', sliceno, temp=False)
	else:
		job.save(data, 'data', sliceno, temp=False)

def verify(params, jwf):
	jid = build('test_jobwithfile', options=dict(inner=True, file=jwf))
	jj = jid.withfile('inner.json', True)
	for sliceno in range(params.slices):
		assert jid.load('inner.pickle', sliceno) == {'inner': sliceno}
		assert jid.json_load('inner.json', sliceno) == {'inner': sliceno}
		assert json.load(jj.open(sliceno=sliceno)) == {'inner': sliceno}
	assert jid.load('inner.pickle') == {'inner': None}
	assert jid.json_load('inner.json') == {'inner': None}
	# use different ways to construct the jwf so both get tested.
	jj = JobWithFile(jid, 'inner.json')
	assert json.load(jj.open()) == {'inner': None}

def synthesis(params, job):
	data = {'foo'}
	if options.inner:
		if options.file.sliced:
			try:
				options.file.load()
				raise Exception("Allowed unsliced load of sliced file")
			except AssertionError:
				pass
		else:
			value = options.file.load()
			assert value == data
			value = blob.load(options.file.filename())
			assert value == data
		job.save({'inner': None}, 'inner.pickle', temp=False)
		job.json_save({'inner': None}, 'inner.json', temp=False)
	else:
		job.save(data, 'data', temp=False)
		# construct these JobWithFile the opposite way to the ones inside verify
		verify(params, JobWithFile(params.jobid, 'data'))
		verify(params, job.withfile('data', True))



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import sys
from os import environ
from os.path import join
from subprocess import Popen, PIPE
import json
from accelerator.compat import unquote_plus


def main(argv, cfg):
	prog = argv.pop(0)
	if argv and argv[0] in ('server', 'urd',):
		which = argv.pop(0)
	else:
		which = 'urd'
	if '--help' in argv or '-h' in argv or not argv:
		fh = sys.stdout if argv else sys.stderr
		print('usage: %s [server|urd] [curl options] path' % (prog,), file=fh)
		print('%s server talks to the server, %s urd talks to urd (default)' % (prog, prog,), file=fh)
		print(file=fh)
		print('examples:', file=fh)
		print('  %s %s/example/latest' % (prog, environ['USER'],), file=fh)
		print('  %s server status' % (prog,), file=fh)
		return
	url_end = argv.pop()
	socket_opts = []
	if which == 'urd':
		url_start = cfg.urd
	else: # server
		url_start = cfg.url
	if url_start.startswith('unixhttp://'):
		url_start = url_start.split('://', 1)[1]
		if '/' in url_start:
			socket, url_start = url_start.split('/', 1)
		else:
			socket, url_start = url_start, ''
		socket_opts = ['--unix-socket', unquote_plus(socket)]
		url_start = join('http://.', url_start)
	argv = ['curl', '-sS'] + socket_opts + argv + [join(url_start, url_end)]
	curl = Popen(argv, stdout=PIPE)
	output, _ = curl.communicate()
	if output:
		try:
			output = output.decode('utf-8')
			output = json.dumps(json.loads(output), indent=4)
		except Exception:
			pass
		print(output)
	return curl.wait()



############################################################################
#                                                                          #
# Copyright (c) 2020-2022 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

import bottle
import json
import sys
import os
import tarfile
import itertools
import collections
import functools

from accelerator.job import Job, JobWithFile
from accelerator.dataset import Dataset
from accelerator.unixhttp import call, WaitressServer
from accelerator.build import fmttime
from accelerator.configfile import resolve_listen
from accelerator.error import NoSuchWhateverError
from accelerator.shell.workdir import job_data, workdir_jids
from accelerator.compat import setproctitle, url_quote, urlencode
from accelerator import __version__ as ax_version

def get_job(jobid):
	if jobid.endswith('-LATEST'):
		base = jobid.rsplit('-', 1)[0]
		jobid = os.readlink(Job(base + '-0').path[:-2] + '-LATEST')
	return Job(jobid)

# why wasn't Accept specified in a sane manner (like sending it in preference order)?
def get_best_accept(*want):
	d = {want[0]: -1} # fallback to first specified
	# {'a/*': 'a/exact'}, reversed() so earlier win
	want_short = {w.split('/', 1)[0] + '/*': w for w in reversed(want)}
	want = set(want)
	want.update(want_short)
	for accept in bottle.request.headers.get('Accept', '').split(','):
		accept = accept.split(';')
		mimetype = accept[0].strip()
		if mimetype in want:
			d[mimetype] = 1.0
			for p in accept[1:]:
				p = p.strip()
				if p.startswith('q='):
					try:
						d[mimetype] = float(p[2:])
					except ValueError:
						pass
	# include k in sort key as well so /* gets lower priority
	_, best = sorted(((v, k) for k, v in d.items()), reverse=True)[0]
	return want_short.get(best, best)

class JSONEncoderWithSet(json.JSONEncoder):
	__slots__ = ()
	def default(self, o):
		if isinstance(o, set):
			return list(o)
		return json.JSONEncoder.default(o)

json_enc = JSONEncoderWithSet(indent=4, ensure_ascii=False).encode


def ax_repr(o):
	res = []
	if isinstance(o, JobWithFile):
		link = '/job/' + bottle.html_escape(o.job)
		res.append('JobWithFile(<a href="%s">job=' % (link,))
		res.append(ax_repr(o.job))
		name = bottle.html_escape(o.name)
		if o.sliced:
			name += '.0'
		res.append('</a>, <a href="%s/%s">name=' % (link, name,))
		res.append(ax_repr(o.name))
		res.append('</a>, sliced=%s, extra=%s' % (ax_repr(o.sliced), ax_repr(o.extra),))
		res.append(')')
	elif isinstance(o, (list, tuple)):
		bra, ket = ('[', ']',) if isinstance(o, list) else ('(', ')',)
		res.append(bra)
		comma = ''
		for v in o:
			res.append(comma)
			res.append(ax_repr(v))
			comma = ', '
		res.append(ket)
	elif isinstance(o, dict):
		res.append('{')
		comma = ''
		for k, v in o.items():
			res.append(comma)
			res.append(ax_repr(k))
			res.append(': ')
			res.append(ax_repr(v))
			comma = ', '
		res.append('}')
	else:
		res.append(bottle.html_escape(repr(o)))
	return ''.join(res)

def ax_link(v):
	if isinstance(v, tuple):
		return '(%s)' % (', '.join(ax_link(vv) for vv in v),)
	elif isinstance(v, list):
		return '[%s]' % (', '.join(ax_link(vv) for vv in v),)
	elif v:
		ev = bottle.html_escape(v)
		if isinstance(v, Dataset):
			job = bottle.html_escape(v.job)
			name = bottle.html_escape(v.name)
			return '<a href="/job/%s">%s</a>/<a href="/dataset/%s">%s</a>' % (job, job, ev, name,)
		elif isinstance(v, Job):
			return '<a href="/job/%s">%s</a>' % (ev, ev,)
		else:
			return ev
	else:
		return ''

def template(tpl_name, **kw):
	return bottle.template(
		tpl_name,
		ax_repr=ax_repr,
		ax_link=ax_link,
		ax_version=ax_version,
		template=template,
		**kw
	)


def view(name, subkey=None):
	def view_decorator(func):
		@functools.wraps(func)
		def view_wrapper(**kw):
			res = func(**kw)
			if isinstance(res, dict):
				accept = get_best_accept('application/json', 'text/json', 'text/html')
				if accept == 'text/html':
					return template(name, **res)
				else:
					bottle.response.content_type = accept + '; charset=UTF-8'
					if callable(subkey):
						res = subkey(res)
					elif subkey:
						res = res[subkey]
					return [json_enc(res), '\n']
			return res
		return view_wrapper
	return view_decorator

def fix_stacks(stacks, report_t):
	pid2pid = {}
	pid2jid = {}
	pid2part = {}
	job_pid = None
	for pid, indent, msg, t in stacks:
		if pid not in pid2pid and pid not in pid2jid:
			if msg.startswith('analysis('):
				pid2part[pid] = ''.join(c for c in msg if c.isdigit())
				pid2pid[pid] = job_pid
			else:
				pid2jid[pid] = msg.split(' ', 1)[0]
				job_pid = pid
		elif pid not in pid2part:
			pid2part[pid] = msg if msg in ('prepare', 'synthesis') else 'analysis'
		jobpid = pid
		while jobpid in pid2pid:
			jobpid = pid2pid[jobpid]
		jid = pid2jid[jobpid]
		if indent < 0:
			msg = msg.split('\n')
			start = len(msg) - 1
			while start and sum(map(bool, msg[start:])) < 5:
				start -= 1
			msg = [line.rstrip('\r') for line in msg[start:]]
			t = fmttime(report_t - t)
		else:
			t = fmttime(report_t - t, short=True)
		yield (jid, pid, indent, pid2part.get(pid), msg, t)

# datasets aren't dicts, so can't be usefully json encoded
def ds_json(d):
	ds = d['ds']
	keys = ('job', 'name', 'parent', 'filename', 'previous', 'hashlabel', 'lines')
	res = {k: getattr(ds, k) for k in keys}
	res['method'] = ds.job.method
	res['columns'] = {k: c.type for k, c in ds.columns.items()}
	return res

def main(argv, cfg):
	prog = argv.pop(0)
	if '-h' in argv or '--help' in argv or len(argv) not in (0, 1):
		print('usage: %s [listen_on]' % (prog,))
		print('runs a web server on listen_on (default localhost:8520, can be socket path)')
		print('for displaying results (result_directory)')
		return
	if argv:
		listen = argv[0]
	else:
		listen = 'localhost:8520'
	cfg.board_listen = resolve_listen(listen)[0]
	if isinstance(cfg.board_listen, str):
		# The listen path may be relative to the directory the user started us
		# from, but the reloader will exec us from the project directory, so we
		# have to be a little gross.
		cfg.board_listen = os.path.join(cfg.user_cwd, cfg.board_listen)
		sys.argv[sys.argv.index(listen)] = cfg.board_listen
	run(cfg, from_shell=True)

def run(cfg, from_shell=False):
	project = os.path.split(cfg.project_directory)[1]
	setproctitle('ax board-server for %s on %s' % (project, cfg.board_listen,))

	def call_s(*path):
		return call(os.path.join(cfg.url, *map(url_quote, path)))

	def call_u(*path, **kw):
		url = os.path.join(cfg.urd, *map(url_quote, path))
		if kw:
			url = url + '?' + urlencode(kw)
		return call(url, server_name='urd')

	@bottle.get('/')
	@view('main')
	def main_page():
		return dict(
			project=project,
			workdirs=cfg.workdirs,
		)

	@bottle.get('/results')
	def results():
		files = {}
		res = {'files': files}
		filenames = ()
		try:
			filenames = os.listdir(cfg.result_directory)
		except KeyError:
			res['missing'] = 'result directory not configured'
		except OSError:
			res['missing'] = 'result directory %r missing' % (cfg.result_directory,)
		for fn in filenames:
			if fn.endswith('_'):
				continue
			ffn = os.path.join(cfg.result_directory, fn)
			try:
				jobid, name = os.readlink(ffn).split('/')[-2:]
				files[fn] = dict(
					jobid=jobid,
					name=name,
					ts=os.lstat(ffn).st_mtime,
					size=os.stat(ffn).st_size,
				)
			except OSError:
				continue
		bottle.response.content_type = 'application/json; charset=UTF-8'
		bottle.response.set_header('Cache-Control', 'no-cache')
		return json.dumps(res)

	@bottle.get('/results/<name>')
	def file(name):
		return bottle.static_file(name, root=cfg.result_directory)

	@bottle.get('/status')
	@view('status')
	def status():
		status = call_s('status/full')
		if 'short' in bottle.request.query:
			if status.idle:
				return 'idle'
			else:
				t, msg, _ = status.current
				return '%s (%s)' % (msg, fmttime(status.report_t - t, short=True),)
		else:
			status.tree = list(fix_stacks(status.pop('status_stacks', ()), status.report_t))
			return status

	@bottle.get('/last_error')
	@view('last_error')
	def last_error():
		return call_s('last_error')

	@bottle.get('/job/<jobid>/method.tar.gz/')
	@bottle.get('/job/<jobid>/method.tar.gz/<name:path>')
	def job_method(jobid, name=None):
		job = get_job(jobid)
		with tarfile.open(job.filename('method.tar.gz'), 'r:gz') as tar:
			if name:
				info = tar.getmember(name)
			else:
				members = [info for info in tar.getmembers() if info.isfile()]
				if len(members) == 1 and not name:
					info = members[0]
				else:
					return template('job_method_list', members=members, job=job)
			bottle.response.content_type = 'text/plain; charset=UTF-8'
			return tar.extractfile(info).read()

	@bottle.get('/job/<jobid>/<name:path>')
	def job_file(jobid, name):
		job = get_job(jobid)
		res = bottle.static_file(name, root=job.path)
		if not res.content_type and res.status_code < 400:
			# bottle default is text/html, which is probably wrong.
			res.content_type = 'text/plain'
		return res

	@bottle.get('/job/<jobid>')
	@bottle.get('/job/<jobid>/')
	@view('job')
	def job(jobid):
		job = get_job(jobid)
		try:
			post = job.post
		except IOError:
			post = None
		if post:
			aborted = False
			files = [fn for fn in job.files() if fn[0] != '/']
			subjobs = [Job(jobid) for jobid in post.subjobs]
			current = call_s('job_is_current', job)
		else:
			aborted = True
			current = False
			files = None
			subjobs = None
		return dict(
			job=job,
			aborted=aborted,
			current=current,
			output=os.path.exists(job.filename('OUTPUT')),
			datasets=job.datasets,
			params=job.params,
			subjobs=subjobs,
			files=files,
		)

	@bottle.get('/dataset/<dsid:path>')
	@view('dataset', ds_json)
	def dataset(dsid):
		ds = Dataset(dsid.rstrip('/'))
		q = bottle.request.query
		if q.column:
			lines = int(q.lines or 10)
			it = ds.iterate(None, q.column)
			it = itertools.islice(it, lines)
			t = ds.columns[q.column].type
			if t in ('datetime', 'date', 'time',):
				it = map(str, it)
			elif t in ('bytes', 'pickle',):
				it = map(repr, it)
			res = list(it)
			bottle.response.content_type = 'application/json; charset=UTF-8'
			return json.dumps(res)
		else:
			return dict(ds=ds)

	def load_workdir(jobs, name):
		known = call_s('workdir', name)
		jobs[name + '-LATEST'] = None # Sorts first
		try:
			latest = os.readlink(os.path.join(cfg.workdirs[name], name + '-LATEST'))
		except OSError:
			latest = None
		for jid in workdir_jids(cfg, name):
			jobs[jid] = job_data(known, jid)
		if latest in jobs:
			jobs[name + '-LATEST'] = jobs[latest]
		else:
			del jobs[name + '-LATEST']
		return jobs

	@bottle.get('/workdir/<name>')
	@view('workdir', 'jobs')
	def workdir(name):
		return dict(name=name, jobs=load_workdir(collections.OrderedDict(), name))

	@bottle.get('/workdir')
	@bottle.get('/workdir/')
	@view('workdir', 'jobs')
	def all_workdirs():
		jobs = collections.OrderedDict()
		for name in sorted(cfg.workdirs):
			load_workdir(jobs, name)
		return dict(name='ALL', jobs=jobs)

	@bottle.get('/methods')
	@view('methods')
	def methods():
		methods = call_s('methods')
		by_package = collections.defaultdict(list)
		for name, data in sorted(methods.items()):
			by_package[data.package].append(name)
		by_package.pop('accelerator.test_methods', None)
		return dict(methods=methods, by_package=by_package)

	@bottle.get('/method/<name>')
	@view('method', 'data')
	def method(name):
		methods = call_s('methods')
		if name not in methods:
			return bottle.HTTPError(404, 'Method %s not found' % (name,))
		return dict(name=name, data=methods[name], cfg=cfg)

	@bottle.get('/urd')
	@bottle.get('/urd/')
	@view('urd', 'lists')
	def urd():
		return dict(
			lists=call_u('list'),
			project=project,
		)

	@bottle.get('/urd/<user>/<build>')
	@bottle.get('/urd/<user>/<build>/')
	@view('urdlist', 'timestamps')
	def urdlist(user, build):
		key = user + '/' + build
		return dict(
			key=key,
			timestamps=call_u(key, 'since/0', captions=1),
		)

	@bottle.get('/urd/<user>/<build>/<ts>')
	@view('urditem', 'entry')
	def urditem(user, build, ts):
		key = user + '/' + build + '/' + ts
		d = call_u(key)
		return dict(key=key, entry=d)

	@bottle.error(500)
	def error(e):
		tpl = bottle.ERROR_PAGE_TEMPLATE
		if isinstance(e.exception, NoSuchWhateverError):
			e.body = str(e.exception)
		else:
			# awful hack: replace DEBUG with something that will be true,
			# so that tracebacks are shown (without needing to turn debug on).
			tpl = tpl.replace('DEBUG', '__version__')
		return bottle.template(tpl, e=e)

	bottle.TEMPLATE_PATH = [os.path.join(os.path.dirname(__file__), 'board')]
	if from_shell:
		kw = {'reloader': True}
	else:
		kw = {'quiet': True}
	kw['server'] = WaitressServer
	listen = cfg.board_listen
	if isinstance(listen, tuple):
		kw['host'], kw['port'] = listen
	else:
		from accelerator.server import check_socket
		check_socket(listen)
		kw['host'] = listen
		kw['port'] = 0
	bottle.run(**kw)



description = "Dataset: Create a dataset with number, string, and json columns."


def prepare(job):
	dw = job.datasetwriter()
	dw.add('Numbercolumn', 'number')
	dw.add('Stringcolumn', 'ascii')
	dw.add('Jsoncolumn', 'json')
	return dw


def analysis(prepare_res, sliceno):
	dw = prepare_res
	data = sliceno + 1000
	dw.write(data, str(data) + '-foo', {'n': data, 's': str(data)})



from os.path import dirname, join
from .printer import prt

description = "Jobs: depend_extra howto"


def main(urd):
	prt.source(__file__)
	prt()
	prt('The following method shows how to do "depend_extra"')
	job = urd.build('example_dependextra')

	prt()
	prt('The job will be re-built if there are changes to')
	prt('"mydependency.txt" or "mydependency.py.')

	prt()
	prt('source code is here:')
	prt.plain(join(dirname(__file__), 'a_%s.py' % (job.method,)))



from . import mydependency

description = 'depend_extra'

# This method will be re-executed if any of
#    the file "mydependencyfile.txt" or
#    the module "mydependency"
# is changed.
depend_extra = ('mydependencyfile.txt', mydependency)


def synthesis():
	pass



description = "Read result from input job, add it to itself and return it."

# "jobs" is a list (or tuple) of jobs assigned by the build script.
# Entries that are left unassigned are set to None.
jobs = ('first',)


def synthesis():
	# All items in the jobs variable will be populated by Job objects
	# when executed.
	# The Job.load() function is used to load the job's return value.
	res = jobs.first.load()
	return res + res



from .printer import prt

description = "Jobs: Build and link two jobs, load and print return values"


def main(urd):
	prt.source(__file__)
	prt()
	prt('Run methods "example_returnhelloworld" + "example_returnconcatinput"')
	job1 = urd.build('example_returnhelloworld')
	job2 = urd.build('example_returnconcatinput', first=job1)  # "first" is an item in the "jobs" list in a_example2.py

	prt()
	prt('Return value from "example_returnhelloworld"')
	prt.output(job1.load())
	prt()
	prt('Return value from "example_returnconcatinput"')
	prt.output(job2.load())



datasets = ('source',)


def analysis(sliceno):
	# sum the test column of the source dataset
	# Faster: "return sum(datasets.source.iterate(slinceo, 'Numbercolumn'))"
	c = 0
	for test in datasets.source.iterate(sliceno, 'Numbercolumn'):
		c += test
	return c


def synthesis(analysis_res):
	return analysis_res.merge_auto()



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2022 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

from traceback import print_exc
from datetime import datetime, date, time, timedelta
import sys

from accelerator.compat import iteritems, itervalues, first_value, str_types, int_types, num_types, unicode

from accelerator.extras import OptionEnum, OptionEnumValue, _OptionString, OptionDefault, RequiredOption, typing_conv
from accelerator.job import JobWithFile

class OptionException(Exception):
	__slots__ = ()
	pass

_date_types = (datetime, date, time, timedelta)

class DepTree:

	def __init__(self, methods, setup):
		self.methods = methods
		m = self.method = setup.method
		self.params = {
			'options': setup.options,
			'datasets': setup.datasets,
			'jobs': setup.jobs,
		}
		self.item = {
			'method': m,
			'params': {m: self.params},
			'make': False,
			'link': False,
			'uid': 0,
		}
		self._fix_options(False)
		self._fix_jobids('jobs')
		self._fix_jobids('datasets')

	def get_reqlist(self):
		params = {k: dict(v) for k, v in iteritems(self.methods.params[self.method].defaults)}
		for k, v in iteritems(self.params):
			params[k].update(v)
		return [(self.method, 0, self.methods.params2optset({self.method: params}))]

	def fill_in_default_options(self):
		self._fix_options(True)

	def _fix_jobids(self, key):
		method = self.method
		data = self.params[key]
		method_wants = self.methods.params[method][key]
		res = {}
		for jobid_name in method_wants:
			if isinstance(jobid_name, str_types):
				value = data.get(jobid_name)
				assert value is None or isinstance(value, str), 'Input %s on %s not a string as required' % (jobid_name, method,)
			elif isinstance(jobid_name, list):
				if len(jobid_name) != 1 or not isinstance(jobid_name[0], str_types):
					raise OptionException('Bad %s item on %s: %s' % (key, method, repr(jobid_name),))
				jobid_name = jobid_name[0]
				value = data.get(jobid_name)
				if value:
					if isinstance(value, str_types):
						value = [e.strip() for e in value.split(',')]
				else:
					value = []
				assert isinstance(value, list), 'Input %s on %s not a list or string as required' % (jobid_name, method,)
			else:
				raise OptionException('%s item of unknown type %s on %s: %s' % (key, type(jobid_name), method, repr(jobid_name),))
			res[jobid_name] = value
		self.params[key] = res
		spill = set(data) - set(res)
		if spill:
			raise OptionException('Unknown %s on %s: %s' % (key, method, ', '.join(sorted(spill)),))

	def _fix_options(self, fill_in):
		method = self.method
		options = self.methods.params[method].options
		res_options = {}
		def typefuzz(t):
			if issubclass(t, str_types):
				return str_types
			if issubclass(t, int_types):
				return int_types
			return t
		def convert(default_v, v):
			if isinstance(default_v, RequiredOption):
				if v is None and not default_v.none_ok:
					raise OptionException('Option %s on method %s requires a non-None value (%r)' % (k, method, default_v.value,))
				default_v = default_v.value
			if default_v is None or v is None:
				if isinstance(default_v, _OptionString):
					raise OptionException('Option %s on method %s requires a non-empty string value' % (k, method,))
				if hasattr(default_v, '_valid') and v not in default_v._valid:
					raise OptionException('Option %s on method %s requires a value in %s' % (k, method, default_v._valid,))
				if isinstance(default_v, OptionDefault):
					v = default_v.default
				return v
			if isinstance(default_v, OptionDefault):
				default_v = default_v.value
			if isinstance(default_v, dict) and isinstance(v, dict):
				if default_v:
					sample_v = first_value(default_v)
					for chk_v in itervalues(default_v):
						assert isinstance(chk_v, type(sample_v))
					return {k: convert(sample_v, v) for k, v in iteritems(v)}
				else:
					return v
			if isinstance(default_v, (list, set, tuple,)) and isinstance(v, str_types + (list, set, tuple,)):
				if isinstance(v, str_types):
					v = (e.strip() for e in v.split(','))
				if default_v:
					sample_v = first_value(default_v)
					for chk_v in default_v:
						assert isinstance(chk_v, type(sample_v))
					v = (convert(sample_v, e) for e in v)
				return type(default_v)(v)
			if isinstance(default_v, (OptionEnum, OptionEnumValue,)):
				if not (v or None) in default_v._valid:
					ok = False
					for cand_prefix in default_v._prefixes:
						if v.startswith(cand_prefix):
							ok = True
							break
					if not ok:
						raise OptionException('%r not a permitted value for option %s on method %s (%s)' % (v, k, method, default_v._valid))
				return v or None
			if isinstance(default_v, str_types + num_types) and isinstance(v, str_types + num_types):
				if isinstance(default_v, _OptionString):
					v = str(v)
					if not v:
						raise OptionException('Option %s on method %s requires a non-empty string value' % (k, method,))
					return v
				if isinstance(default_v, unicode) and isinstance(v, bytes):
					return v.decode('utf-8')
				return type(default_v)(v)
			if (isinstance(default_v, type) and isinstance(v, typefuzz(default_v))) or isinstance(v, typefuzz(type(default_v))):
				return v
			if isinstance(default_v, bool) and isinstance(v, (str, int)):
				lv = str(v).lower()
				if lv in ('true', '1', 't', 'yes', 'on',):
					return True
				if lv in ('false', '0', 'f', 'no', 'off', '',):
					return False
			if isinstance(default_v, _date_types):
				default_v = type(default_v)
			if default_v in _date_types:
				try:
					return typing_conv[default_v.__name__](v)
				except Exception:
					raise OptionException('Failed to convert option %s %r to %s on method %s' % (k, v, default_v, method,))
			if isinstance(v, str_types) and not v:
				return type(default_v)()
			if isinstance(default_v, JobWithFile) or default_v is JobWithFile:
				defaults = ('', '', False, None,)
				if default_v is JobWithFile:
					default_v = defaults
				if not isinstance(v, (list, tuple,)) or not (2 <= len(v) <= 4):
					raise OptionException('Option %s (%r) on method %s is not %s compatible' % (k, v, method, type(default_v)))
				v = tuple(v) + defaults[len(v):] # so all of default_v gets convert()ed.
				v = [convert(dv, vv) for dv, vv in zip(default_v, v)]
				return JobWithFile(*v)
			if type(default_v) != type:
				default_v = type(default_v)
			raise OptionException('Failed to convert option %s of %s to %s on method %s' % (k, type(v), default_v, method,))
		for k, v in iteritems(self.params['options']):
			if k in options:
				try:
					res_options[k] = convert(options[k], v)
				except OptionException:
					raise
				except Exception:
					print_exc(file=sys.stderr)
					raise OptionException('Failed to convert option %s on method %s' % (k, method,))
			else:
				raise OptionException('Unknown option %s on method %s' % (k, method,))
		if fill_in:
			missing = set(options) - set(res_options)
			missing_required = missing & self.methods.params[method].required
			if missing_required:
				raise OptionException('Missing required options {%s} on method %s' % (', '.join(sorted(missing_required)), method,))
			defaults = self.methods.params[method].defaults
			res_options.update({k: defaults.options[k] for k in missing})
		self.params['options'] = res_options

	def set_link(self, uid, job):
		self.item['link'] = job.id
		self.item['total_time'] = job.total

	def propagate_make(self):
		self.item['make'] = not self.item['link']

	def get_sorted_joblist(self):
		return [self.item]



from collections import Counter

description = "Dataset: Iterate and aggregate a column"

datasets = ('source',)


def analysis(sliceno):
	v = Counter()
	for s, n in datasets.source.iterate(sliceno, ['String', 'Int']):
		v[s] += n
	return v


def synthesis(analysis_res):
	v = analysis_res.merge_auto()  # Don't merge or sort huge datasets, though!
	for item in sorted(v.items()):
		print('print some stuff', item)
	return v



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test re-using datasets (from test_selfchain) in new
chains, and verify that the old chain still works.
Also tests that the dataset cache is updated correctly
on re-chaining.
'''

from accelerator.dataset import Dataset, DatasetWriter

jobs = ('selfchain',)

def synthesis(job):
	manual_chain = [Dataset(jobs.selfchain, name) for name in "abcdefgh"]
	manual_abf = [manual_chain[0], manual_chain[1], manual_chain[5]]
	# build a local abf chain
	prev = None
	for ix, ds in enumerate(manual_abf):
		name = "abf%d" % (ix,)
		prev = ds.link_to_here(name, override_previous=prev)
	manual_abf_data = list(Dataset.iterate_list(None, None, manual_abf))
	local_abf_data = list(Dataset(job, "abf2").iterate_chain(None, None))
	assert manual_abf_data == local_abf_data
	# disconnect h, verify there is no chain
	manual_chain[-1].link_to_here("alone", override_previous=None)
	assert len(Dataset(job, "alone").chain()) == 1
	# check that the original chain is unhurt
	assert manual_chain == manual_chain[-1].chain()

	# So far so good, now make a chain long enough to have a cache.
	prev = None
	ix = 0
	going = True
	while going:
		if prev and "cache" in prev._data:
			going = False
		name = "longchain%d" % (ix,)
		dw = DatasetWriter(name=name, previous=prev)
		dw.add("ix", "number")
		dw.get_split_write()(ix)
		prev = dw.finish()
		ix += 1
	# we now have a chain that goes one past the first cache point
	full_chain = Dataset(prev).chain()
	assert "cache" in full_chain[-2]._data # just to check the above logic is correct
	assert "cache" not in full_chain[-1]._data # just to be sure..
	full_chain[-2].link_to_here("nocache", override_previous=None)
	full_chain[-1].link_to_here("withcache", override_previous=full_chain[-3])
	assert "cache" not in Dataset(job, "nocache")._data
	assert "cache" in Dataset(job, "withcache")._data
	# And make sure they both get the right data too.
	assert list(Dataset(prev).iterate_chain(None, "ix")) == list(range(ix))
	assert list(Dataset(job, "nocache").iterate_chain(None, "ix")) == [ix - 2]
	assert list(Dataset(job, "withcache").iterate_chain(None, "ix")) == list(range(ix - 2)) + [ix - 1]



description = "Dataset: Append a column to an existing dataset."

datasets = ('source',)


def prepare(job):
	dw = job.datasetwriter(parent=datasets.source)
	dw.add('Floatcolumn', 'float64')
	return dw


def analysis(prepare_res, sliceno):
	dw = prepare_res
	for n in datasets.source.iterate(sliceno, 'Numbercolumn'):
		dw.write(n * 3.14)



############################################################################
#                                                                          #
# Copyright (c) 2019-2021 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify the various dataset_type chaining options:
Building a chain with and without extra datasets in the source.
Using the length option to get only some of the extra datasets.
Using the as_chain option to get one dataset per slice.
'''

from accelerator.dataset import Dataset
from accelerator.extras import DotDict
from accelerator import subjobs

def synthesis(job, slices):
	def verify(a, b):
		for col in 'abcd':
			for sliceno in range(slices):
				a_data = list(Dataset.iterate_list(sliceno, col, a))
				b_data = list(map(str, Dataset.iterate_list(sliceno, col, b)))
				assert a_data == b_data, '%r has different contents to %r in slice %d column %s' % (a, b, sliceno, col,)
	def verify_sorted(a, b):
		for col in 'abcd':
			a_data = list(Dataset.iterate_list(None, col, a))
			b_data = list(map(str, Dataset.iterate_list(None, col, b)))
			a_data.sort()
			b_data.sort()
			assert a_data == b_data, '%r has different contents to %r in column %s' % (a, b, col,)
	def write(name, previous, low, high, filter=lambda ix: True):
		dw = job.datasetwriter(
			name=name,
			previous=previous,
			columns={'a': 'unicode', 'b': 'unicode', 'c': 'unicode', 'd': 'unicode',},
		)
		w = dw.get_split_write()
		for ix in range(low, high):
			if filter(ix):
				w('%d' % (ix,), '%d.2' % (ix,), '%d%s' % (ix, '.5' if ix % 2 else ''), '[%d]' % (ix,))
		return dw.finish()
	untyped_A = write('A', None, 0, 100)
	untyped_B = write('B', untyped_A, 100, 1000)
	untyped_C = write('C', untyped_B, 1000, 2000)
	untyped_D = write('D', untyped_C, 2000, 10000)
	untyped_E = write('E', untyped_D, 10000, 10100)

	# All four different classes of converters
	opts = DotDict(column2type=dict(a='int32_10', b='number', c='ascii', d='json'), as_chain=False)
	src_chain = []
	simple_chain = []
	previous = None
	for src in (untyped_A, untyped_B, untyped_C, untyped_D, untyped_E):
		previous = subjobs.build('dataset_type', datasets=dict(source=src, previous=previous), options=opts)
		simple_chain.append(previous)
		src_chain.append(src)
		verify([src], [previous])
		assert simple_chain == Dataset(previous).chain(), previous
		verify(src_chain, simple_chain)
	typed_B = simple_chain[1]
	typed_D = simple_chain[3]

	# No previous -> should contain both A and B
	typed_AB = subjobs.build('dataset_type', datasets=dict(source=untyped_B), options=opts)
	verify(src_chain[:2], [typed_AB])
	typed_CDE = subjobs.build('dataset_type', datasets=dict(source=untyped_E, previous=typed_B), options=opts)
	verify(src_chain[2:], [typed_CDE])
	verify(src_chain, Dataset(typed_CDE).chain())
	# A and B through typed_B, but length=2 only gets D and E, not C.
	opts.length = 2
	typed_DE_noC = subjobs.build('dataset_type', datasets=dict(source=untyped_E, previous=typed_B), options=opts)
	del opts.length
	verify((untyped_A, untyped_B, untyped_D, untyped_E), Dataset(typed_DE_noC).chain())

	# with as_chain (and a hashlabel so as_chain happens)
	opts.as_chain = True
	opts.hashlabel = 'a'
	previous = None
	used_slices = []
	for ix, src in enumerate(src_chain, 1):
		used_slices.append(min(slices, sum(src.lines)))
		previous = subjobs.build('dataset_type', datasets=dict(source=src, previous=previous), options=opts)
		ds = Dataset(previous)
		assert len(ds.chain()) == sum(used_slices), ds
		verify_sorted([src], ds.chain(length=used_slices[-1]))
		verify_sorted(src_chain[:ix], ds.chain())

	# And one with as_chain just on the last job, discarding half the rows from bad typing.
	opts.column2type['b'] = 'ascii'
	opts.column2type['c'] = 'number:int'
	opts.filter_bad = True
	typed_and_hashed_Ehalf = subjobs.build('dataset_type', datasets=dict(source=untyped_E, previous=typed_D), options=opts)
	typed_and_hashed_Ehalf_bad = Dataset(typed_and_hashed_Ehalf, 'bad')
	assert len(typed_and_hashed_Ehalf_bad.chain()) == 1, typed_and_hashed_Ehalf_bad
	assert set(typed_and_hashed_Ehalf_bad.columns) == set(opts.column2type), typed_and_hashed_Ehalf_bad
	assert sum(typed_and_hashed_Ehalf_bad.lines) == 50, typed_and_hashed_Ehalf_bad
	typed_and_hashed_Ehalf = Dataset(typed_and_hashed_Ehalf)
	assert len(typed_and_hashed_Ehalf.chain()) == used_slices[-1] + 4, typed_and_hashed_Ehalf
	untyped_Ehalf = write('Ehalf', untyped_D, 10000, 10100, filter=lambda ix: ix % 2 == 0)
	verify_sorted([untyped_Ehalf], typed_and_hashed_Ehalf.chain(length=used_slices[-1]))



from os.path import dirname, join
from .printer import prt

description = "Dataset: Create a chained dataset."

# Files are stored in same directory as this python file,
# see comment below.
path = dirname(__file__)


def main(urd):
	prt.source(__file__)
	prt()
	prt('Create a chain of datasets using csvimport.')
	imp = None
	for filename in ('data.csv', 'data2.csv', 'data3.csv'):
		# Ideally, you'd set "input_directory" to the location of the
		# input files in "accelerator.conf" to avoid an absolute path
		# in the input filename.
		filename = join(path, filename)
		imp = urd.build('csvimport', filename=filename, previous=imp)

	prt()
	prt('Try this to investigate the chain.')
	prt.command('ax ds -c -S', imp)

	prt()
	prt('To go back in chain and investigate datasets, try')
	prt.command('ax ds %s' % (imp,))
	prt.command('ax ds %s~' % (imp,))
	prt.command('ax ds %s~~' % (imp,))
	prt('Note that ~~ can also be written ~2 etc.')

	prt()
	prt('This method will iterate over the whole chain.')
	job = urd.build('dsexample_iteratechain', source=imp)

	prt()
	prt('To see its output, try')
	prt.command('ax job -O', job)



############################################################################
#                                                                          #
# Copyright (c) 2019-2020 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import sys
import hashlib
from importlib import import_module
from collections import namedtuple

from accelerator.compat import unicode, str_types


_prologue_code_template = r'''
/* This is autogenerated from %s.py */

#define PY_SSIZE_T_CLEAN 1
#include <Python.h>
#include <stdint.h>

static PyObject *nullmarker = 0;

static int str_or_0(PyObject *obj, const char **res) {
	if (obj == nullmarker) {
		*res = 0;
		return 0;
	}
	*res = PyBytes_AsString(obj);
	return !*res;
}

static PyObject *py_set_null(PyObject *dummy, PyObject *obj)
{
	Py_CLEAR(nullmarker);
	Py_INCREF(obj);
	nullmarker = obj;
	Py_RETURN_NONE;
}
'''

_init_code_template = r'''
static PyMethodDef module_methods[] = {
	{"set_null", py_set_null, METH_O, 0},
	%(methods)s,
	{NULL, NULL, 0, NULL}
};

#if PY_MAJOR_VERSION < 3
#  define INITERR
#else
#  define INITERR 0
static struct PyModuleDef moduledef = {
	PyModuleDef_HEAD_INIT,
	"_%(name)s",        /*m_name*/
	0,                  /*m_doc*/
	-1,                 /*m_size*/
	module_methods,     /*m_methods*/
	0,                  /*m_reload*/
	0,                  /*m_traverse*/
	0,                  /*m_clear*/
	0,                  /*m_free*/
};
#endif

#if PY_MAJOR_VERSION >= 3
__attribute__ ((visibility("default"))) PyMODINIT_FUNC PyInit__%(name)s(void)
{
	PyObject *m = PyModule_Create(&moduledef);
	if (!m) return 0;
	PyModule_AddObject(m, "source_hash", PyUnicode_FromString(source_hash));
	return m;
}
#else
__attribute__ ((visibility("default"))) PyMODINIT_FUNC init_%(name)s(void)
{
	PyObject *m = Py_InitModule3("_%(name)s", module_methods, NULL);
	if (!m) return;
	PyModule_AddObject(m, "source_hash", PyUnicode_FromString(source_hash));
}
#endif
'''

_method_def_template = r'''{"%s", py_%s, METH_VARARGS, 0}'''

CStuff = namedtuple('CStuff', 'backend NULL mk_uint64 bytesargs')

def make_source(name, functions, protos, extra_functions, extra_method_defs, wrapper_template):
	code = []
	code.append(_prologue_code_template % (name,))
	code.append(functions)
	code.append(extra_functions)
	method_defs = []
	for p in protos:
		funcname = p[11:].split('(')[0]
		code.append(wrapper_template % (funcname, funcname,))
		method_defs.append(_method_def_template % (funcname, funcname,))
	method_defs.extend(extra_method_defs)
	code.append(_init_code_template % dict(methods=',\n\t'.join(method_defs), name=name,))
	hash = hashlib.sha1(b''.join(c.encode('ascii') for c in code)).hexdigest()
	code.insert(-1, 'static char source_hash[] = "%s";\n' % (hash,))
	code = ''.join(code)
	return code, hash

def init(name, hash, protos, extra_protos, functions):
	backend = import_module('accelerator.standard_methods._' + name)
	if hash == backend.source_hash:
		NULL = object()
		backend.set_null(NULL)
		def mk_uint64(count=1):
			return [0] * count
		def str2c(s):
			if isinstance(s, unicode):
				s = s.encode('utf-8')
			return s
	else:
		print('[%s] Backend modified since module was compiled, falling back to cffi for development.' % (name,), file=sys.stderr)
		import cffi
		ffi = cffi.FFI()

		NULL = ffi.NULL
		def mk_uint64(count=1):
			return ffi.new('uint64_t []', [0] * count)
		def str2c(s):
			if isinstance(s, unicode):
				s = s.encode('utf-8')
			return ffi.new('char []', s)

		# cffi apparently doesn't know about off_t.
		# "typedef int... off_t" isn't allowed with verify,
		# so I guess we'll just assume that ssize_t is the same as off_t.
		ffi.cdef('typedef ssize_t off_t;')
		ffi.cdef(''.join(protos + extra_protos))
		backend = ffi.verify(functions, libraries=['z'], extra_compile_args=['-std=c99', '-DCFFI_ATE_MY_GIL'])

	# make any unicode args bytes, for C calls.
	def bytesargs(*a):
		return [
			str2c(v) if isinstance(v, str_types)
			else bytesargs(*v) if isinstance(v, list) and v and isinstance(v[0], str_types)
			else NULL if v is None
			else v
			for v in a
		]

	return CStuff(backend, NULL, mk_uint64, bytesargs)
